#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import itertools
import inspect
import traceback
import threading
from collections import OrderedDict
from datetime import datetime
from decimal import Decimal

from ...compat import six
from ...models import FileResource, TableResource
from .. import types

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


def add_method(expr, methods):
    for k, v in six.iteritems(methods):
        setattr(expr, k, v)


def same_data_source(*exprs):
    curr_data_source = None
    for expr in exprs:
        data_source = sorted(list(expr.data_source()))
        if curr_data_source is None:
            curr_data_source = data_source
        else:
            if curr_data_source != data_source:
                return False

    return True


def highest_precedence_data_type(*data_types):
    data_types = set(data_types)
    if len(data_types) == 1:
        return data_types.pop()

    precedences = dict((t, idx) for idx, t in enumerate(
        [types.string, types.boolean, types.int8, types.int16, types.int32, types.int64,
         types.datetime, types.decimal, types.float32, types.float64]))

    type_precedences = [(precedences[data_type], data_type) for data_type in data_types]
    highest_data_type = sorted(type_precedences)[-1][1]
    for data_type in data_types:
        if data_type != highest_data_type and not highest_data_type.can_implicit_cast(data_type):
            raise TypeError(
                'Type cast error: %s cannot implicitly cast to %s' % (data_type, highest_data_type))

    return highest_data_type


def get_attrs(node):
    from .core import Node

    tp = type(node) if not inspect.isclass(node) else node

    if inspect.getmro(tp) is None:
        tp = type(tp)

    return tuple(OrderedDict.fromkeys(
        it for it in
        itertools.chain(*(cls.__slots__ for cls in inspect.getmro(tp) if issubclass(cls, Node)))
        if not it.startswith('__')))


def get_collection_resources(resources):
    from .expressions import CollectionExpr

    if resources:
        for res in resources:
            if not isinstance(res, (TableResource, FileResource, CollectionExpr)):
                raise ValueError('resources must be ODPS file or table Resources or collections')
    if resources is not None and len(resources) > 0:
        ret = [res for res in resources if isinstance(res, CollectionExpr)]
        [r.cache() for r in ret]  # we should execute the expressions by setting cache=True
        return ret


def get_executed_collection_project_table_name(collection):
    from .expressions import CollectionExpr
    from ...models import Table
    from ..backends.context import context

    if not isinstance(collection, CollectionExpr):
        return

    if collection._source_data is not None and \
            isinstance(collection._source_data, Table):
        source_data = collection._source_data
        return source_data.project.name + '.' + source_data.name

    if context.is_cached(collection) and \
            isinstance(context.get_cached(collection), Table):
        source_data = context.get_cached(collection)
        return source_data.project.name + '.' + source_data.name


def is_called_by_inspector():
    return any(1 for v in traceback.extract_stack() if 'oinspect' in v[0].lower() and 'ipython' in v[0].lower())


def to_list(field):
    if isinstance(field, six.string_types):
        return [field, ]
    if isinstance(field, Iterable):
        return list(field)
    return [field, ]


_lock = threading.Lock()
_index = itertools.count(1)


def new_id():
    with _lock:
        return next(_index)


def select_fields(collection):
    from .expressions import ProjectCollectionExpr, Summary
    from .collections import DistinctCollectionExpr, RowAppliedCollectionExpr
    from .groupby import GroupByCollectionExpr, MutateCollectionExpr

    if isinstance(collection, (ProjectCollectionExpr, Summary)):
        return collection.fields
    elif isinstance(collection, DistinctCollectionExpr):
        return collection.unique_fields
    elif isinstance(collection, (GroupByCollectionExpr, MutateCollectionExpr)):
        return collection.fields
    elif isinstance(collection, RowAppliedCollectionExpr):
        return collection.fields


def is_changed(collection, column):
    # if the column is changed before the generated collection
    from .expressions import CollectionExpr, Column

    column_name = column.source_name
    src_collection = column.input

    if src_collection is collection:
        return False

    dag = collection.to_dag(copy=False, validate=False)
    coll = src_collection
    colls = []
    while coll is not collection:
        try:
            parents = [p for p in dag.successors(coll) if isinstance(p, CollectionExpr)]
        except KeyError:
            return
        assert len(parents) == 1
        coll = parents[0]
        colls.append(coll)

    name = column_name
    for coll in colls:
        fields = select_fields(coll)
        if fields:
            col_names = dict((field.source_name, field) for field in fields if isinstance(field, Column))
            if name in col_names:
                name = col_names[name].name
            else:
                return True

    return False


annotation_rtypes = {
    int: types.int64,
    str: types.string,
    float: types.float64,
    bool: types.boolean,
    datetime: types.datetime,
    Decimal: types.decimal,
}


def get_annotation_rtype(func):
    if hasattr(func, '__annotations__'):
        try:
            from typing import Union
        except ImportError:
            Union = None

        ret_type = func.__annotations__.get('return')
        if ret_type in annotation_rtypes:
            return annotation_rtypes.get(ret_type)
        elif hasattr(ret_type, '__origin__') and ret_type.__origin__ is Union:
            actual_types = [typo for typo in ret_type.__args__
                            if typo is not type(None)]
            if len(actual_types) == 1:
                return annotation_rtypes.get(actual_types[0])
        elif Union is not None and type(ret_type) is type(Union):
            actual_types = [typo for typo in ret_type.__args__
                            if typo is not type(None)]
            if len(actual_types) == 1:
                return annotation_rtypes.get(actual_types[0])
    return None


def get_proxied_expr(expr):
    try:
        obj = object.__getattribute__(expr, '_proxy')
        return obj if obj is not None else expr
    except AttributeError:
        return expr



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import threading
import contextlib

from sqlalchemy import types as sa_types
from sqlalchemy.engine import default, Engine
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.sql import compiler, sqltypes
try:
    from sqlalchemy.dialects import mysql
except ImportError:
    # for low sqlalchemy versions
    from sqlalchemy.databases import mysql

from . import options
from . import types
from .compat import six
from .core import ODPS, DEFAULT_ENDPOINT
from .models import Table
from .models.session import PUBLIC_SESSION_NAME
from .errors import NoSuchObject
from .utils import to_str, to_text


test_setting = threading.local()
test_setting.get_tables_filter = None


@contextlib.contextmanager
def update_test_setting(**kw):
    old_values = {}
    for k in kw:
        old_values[k] = getattr(test_setting, k)

    for k, v in six.iteritems(kw):
        setattr(test_setting, k, v)

    yield

    # set back value
    for k, v in six.iteritems(old_values):
        setattr(test_setting, k, v)


_odps_type_to_sqlalchemy_type = {
    types.Boolean: sa_types.Boolean,
    types.Tinyint: mysql.TINYINT,
    types.Smallint: sa_types.SmallInteger,
    types.Int: sa_types.Integer,
    types.Bigint: sa_types.BigInteger,
    types.Float: sa_types.Float,
    types.Double: sa_types.Float,
    types.String: sa_types.String,
    types.Varchar: sa_types.String,
    types.Char: sa_types.String,
    types.Date: sa_types.Date,
    types.Datetime: sa_types.DateTime,
    types.Timestamp: sa_types.TIMESTAMP,
    types.Binary: sa_types.String,
    types.Array: sa_types.String,
    types.Map: sa_types.String,
    types.Struct: sa_types.String,
    types.Decimal: sa_types.DECIMAL,
}

_sqlalchemy_global_reusable_odps = {}


class ODPSIdentifierPreparer(compiler.IdentifierPreparer):
    # Just quote everything to make things simpler / easier to upgrade
    reserved_words = compiler.RESERVED_WORDS.copy()
    keywords = [
        'ADD', 'ALL', 'ALTER', 'AND', 'AS', 'ASC', 'BETWEEN', 'BIGINT',
        'BOOLEAN', 'BY', 'CASE', 'CAST', 'COLUMN', 'COMMENT', 'CREATE',
        'DESC', 'DISTINCT', 'DISTRIBUTE', 'DOUBLE', 'DROP', 'ELSE', 'FALSE',
        'FROM', 'FULL', 'GROUP', 'IF', 'IN', 'INSERT', 'INTO', 'IS', 'JOIN',
        'LEFT', 'LIFECYCLE', 'LIKE', 'LIMIT', 'MAPJOIN', 'NOT', 'NULL',
        'ON', 'OR', 'ORDER', 'OUTER', 'OVERWRITE', 'PARTITION', 'RENAME',
        'REPLACE', 'RIGHT', 'RLIKE', 'SELECT', 'SORT', 'STRING', 'TABLE',
        'TABLESAMPLE', 'TBLPROPERTIES', 'THEN', 'TOUCH', 'TRUE', 'UNION',
        'VIEW', 'WHEN', 'WHERE'
    ]
    reserved_words.update(keywords)
    reserved_words.update([s.lower() for s in keywords])

    def __init__(self, dialect):
        super(ODPSIdentifierPreparer, self).__init__(
            dialect,
            initial_quote='`',
            escape_quote='`',
        )

    def quote(self, ident, force=None):
        return to_str(super(ODPSIdentifierPreparer, self).quote(ident, force=force))


class ODPSCompiler(compiler.SQLCompiler):
    def visit_column(self, *args, **kwargs):
        result = super(ODPSCompiler, self).visit_column(*args, **kwargs)
        dot_count = result.count('.')
        assert dot_count in (0, 1, 2), "Unexpected visit_column result {}".format(result)
        if dot_count == 2:
            # we have something of the form schema.table.column
            # hive doesn't like the schema in front, so chop it out
            result = result[result.index('.') + 1:]
        return result

    def visit_char_length_func(self, fn, **kw):
        return 'length{}'.format(self.function_argspec(fn, **kw))

    def __unicode__(self):
        return to_text(self)


class ODPSTypeCompiler(compiler.GenericTypeCompiler):
    def visit_INTEGER(self, type_):
        return 'INT'

    def visit_NUMERIC(self, type_):
        return 'DECIMAL'

    def visit_CHAR(self, type_):
        return 'STRING'

    def visit_VARCHAR(self, type_):
        return 'STRING'

    def visit_NCHAR(self, type_):
        return 'STRING'

    def visit_TEXT(self, type_):
        return 'STRING'

    def visit_CLOB(self, type_):
        return 'STRING'

    def visit_BLOB(self, type_):
        return 'BINARY'

    def visit_TIME(self, type_):
        return 'TIMESTAMP'


if hasattr(sqltypes.String, "RETURNS_UNICODE"):
    _return_unicode_str = sqltypes.String.RETURNS_UNICODE
else:
    _return_unicode_str = True


class ODPSDialect(default.DefaultDialect):
    name = 'odps'
    driver = 'rest'
    preparer = ODPSIdentifierPreparer
    statement_compiler = ODPSCompiler
    supports_views = True
    supports_alter = True
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_native_decimal = True
    supports_native_boolean = True
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = _return_unicode_str
    description_encoding = None
    supports_multivalues_insert = True
    type_compiler = ODPSTypeCompiler
    supports_sane_rowcount = False
    supports_statement_cache = False
    _reused_odps = None
    default_schema_name = "default"

    @classmethod
    def dbapi(cls):
        from . import dbapi
        return dbapi

    def create_connect_args(self, url):
        url_string = str(url)
        project = url.host
        if project is None and options.default_project:
            project = options.default_project
        access_id = url.username
        if access_id is None and options.account is not None:
            access_id = options.account.access_id
        secret_access_key = url.password
        if secret_access_key is None and options.account is not None:
            secret_access_key = options.account.secret_access_key
        logview_host = options.logview_host
        endpoint = None
        session_name = None
        use_sqa = False
        reuse_odps = False
        fallback_policy = ''
        hints = {}
        if url.query:
            query = dict(url.query)
            if endpoint is None:
                endpoint = query.pop('endpoint', None)
            if logview_host is None:
                logview_host = query.pop(
                    'logview_host', query.pop('logview', None)
                )
            if session_name is None:
                session_name = query.pop('session', None)
            if use_sqa == False:
                use_sqa = (query.pop('interactive_mode', 'false') != 'false')
            if reuse_odps == False:
                reuse_odps = (query.pop('reuse_odps', 'false') != 'false')
            if fallback_policy == "":
                fallback_policy = query.pop('fallback_policy', 'default')
            hints = query

        if endpoint is None:
            endpoint = options.endpoint or DEFAULT_ENDPOINT
        if session_name is None:
            session_name = PUBLIC_SESSION_NAME

        kwargs = {
            'access_id': access_id,
            'secret_access_key': secret_access_key,
            'project': project,
            'endpoint': endpoint,
            'session_name': session_name,
            'use_sqa': use_sqa,
            'fallback_policy': fallback_policy,
            'hints': hints,
        }
        for k, v in six.iteritems(kwargs):
            if v is None:
                raise ValueError('{} should be provided to create connection, '
                                 'you can either specify in connection string as format: '
                                 '"odps://<access_id>:<access_key>@<project_name>", '
                                 'or create an ODPS object and call `.to_global()` '
                                 'to set it to global'.format(k))
        if logview_host is not None:
            kwargs['logview_host'] = logview_host

        if reuse_odps:
            # the odps object can only be reused only if it will be identical
            if (
                url_string in _sqlalchemy_global_reusable_odps
                and _sqlalchemy_global_reusable_odps.get(url_string) is not None
            ):
                kwargs['odps'] = _sqlalchemy_global_reusable_odps.get(url_string)
                kwargs['access_id'] = None
                kwargs['secret_access_key'] = None
            else:
                _sqlalchemy_global_reusable_odps[url_string] = ODPS(
                    access_id=access_id,
                    secret_access_key=secret_access_key,
                    project=project,
                    endpoint=endpoint,
                    logview_host=logview_host,
                )

        return [], kwargs

    def get_odps_from_url(self, url):
        _, kwargs = self.create_connect_args(url)
        if "odps" in kwargs:
            return kwargs["odps"]
        odps_kw = kwargs.copy()
        odps_kw.pop("session_name", None)
        odps_kw.pop("use_sqa", None)
        odps_kw.pop("fallback_policy", None)
        odps_kw.pop("hints", None)
        odps_kw["overwrite_global"] = False
        return ODPS(**odps_kw)

    def get_schema_names(self, connection, **kw):
        conn = self._get_dbapi_connection(connection)
        if getattr(conn, "_project_as_schema", False):
            fields = ['owner', 'user', 'group', 'prefix']
            if (conn.odps.project is None) or (kw.pop('listall', None) is not None):
                kwargs = {f: kw.get(f) for f in fields}
                return [proj.name for proj in conn.odps.list_projects(**kwargs)]
            else:
                return [conn.odps.project]
        else:
            try:
                return [schema.name for schema in conn.odps.list_schemas()]
            except:
                return ["default"]

    def has_table(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        schema_kw = self._get_schema_kw(connection, schema=schema)
        return conn.odps.exist_table(table_name, **schema_kw)

    @classmethod
    def _get_dbapi_connection(cls, sa_connection):
        if isinstance(sa_connection, Engine):
            sa_connection = sa_connection.connect()
        return sa_connection.connection.connection

    @classmethod
    def _get_schema_kw(cls, connection, schema=None):
        db_conn = cls._get_dbapi_connection(connection)
        if getattr(db_conn, "_project_as_schema", False):
            return dict(project=schema)
        else:
            return dict(schema=schema)

    def get_columns(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        schema_kw = self._get_schema_kw(connection, schema=schema)
        table = conn.odps.get_table(table_name, **schema_kw)
        result = []
        try:
            for col in table.table_schema.columns:
                col_type = _odps_type_to_sqlalchemy_type[type(col.type)]
                result.append({
                    'name': col.name,
                    'type': col_type,
                    'nullable': True,
                    'default': None,
                    'comment': col.comment,
                })
        except NoSuchObject as e:
            # convert ODPSError to SQLAlchemy NoSuchTableError
            raise NoSuchTableError(str(e))
        return result

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        # ODPS has no support for foreign keys.
        return []

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        # ODPS has no support for primary keys.
        return []

    def get_indexes(self, connection, table_name, schema=None, **kw):
        # ODPS has no support for indexes
        return []

    def _iter_tables(self, connection, schema=None, types=None, **kw):
        conn = self._get_dbapi_connection(connection)
        filter_ = getattr(test_setting, 'get_tables_filter', None)
        if filter_ is None:
            filter_ = lambda x: True
        schema_kw = self._get_schema_kw(connection, schema=schema)

        if not types:
            it = conn.odps.list_tables(**schema_kw)
        else:
            its = []
            for table_type in types:
                list_kw = schema_kw.copy()
                list_kw["type"] = table_type
                its.append(conn.odps.list_tables(**list_kw))
            it = itertools.chain(*its)

        return [t.name for t in it if filter_(t.name)]

    def get_table_names(self, connection, schema=None, **kw):
        return self._iter_tables(
            connection,
            schema=schema,
            types=[Table.Type.MANAGED_TABLE, Table.Type.EXTERNAL_TABLE],
            **kw
        )

    def get_view_names(self, connection, schema=None, **kw):
        return self._iter_tables(
            connection,
            schema=schema,
            types=[Table.Type.VIRTUAL_VIEW, Table.Type.MATERIALIZED_VIEW],
            **kw
        )

    def get_table_comment(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        schema_kw = self._get_schema_kw(connection, schema=schema)
        comment = conn.odps.get_table(table_name, **schema_kw).comment
        return {
            'text': comment
        }

    def do_rollback(self, dbapi_connection):
        # No transactions for ODPS
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        # We decode everything as UTF-8
        return True

    def _check_unicode_description(self, connection):
        # We decode everything as UTF-8
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import LazyLoad
from ... import serializers, utils
from ...compat import urlparse


class OfflineModelInfo(serializers.XMLSerializableModel):
    _root = 'Offlinemodel'

    name = serializers.XMLNodeField('Name', default='')
    model_path = serializers.XMLNodeField('ModelPath', default='')
    role_arn = serializers.XMLNodeField('Rolearn')
    type = serializers.XMLNodeField('Type')
    version = serializers.XMLNodeField('Version')
    processor = serializers.XMLNodeField('Processor')
    configuration = serializers.XMLNodeField('Configuration')
    src_project = serializers.XMLNodeField('SrcProject')
    src_model = serializers.XMLNodeField('SrcModel')
    dest_project = serializers.XMLNodeField('DestProject')
    dest_model = serializers.XMLNodeField('DestModel')


class OfflineModel(LazyLoad):
    """
    Representing an ODPS offline model.
    """

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    creation_time = serializers.XMLNodeField(
        'CreationTime', parse_callback=utils.parse_rfc822)
    last_modified_time = serializers.XMLNodeField(
        'LastModifiedTime', parse_callback=utils.parse_rfc822)

    def reload(self):
        resp = self._client.get(self.resource())
        self.parse(self._client, resp, obj=self)

    def get_model(self):
        """
        Get PMML text of the current model. Note that model file obtained
        via this method might be incomplete due to size limitations.
        """
        url = self.resource()
        resp = self._client.get(url, action='data')

        return resp.text

    @utils.with_wait_argument
    def copy(self, new_name, new_project=None, async_=False):
        """
        Copy current model into a new location.

        :param new_name: name of the new model
        :param new_project: new project name. if absent, original project name will be used
        :param async_: if True, return the copy instance. otherwise return the newly-copied model
        """
        url = self.parent.resource()
        new_project = new_project or self.project.name

        info = OfflineModelInfo(src_model=self.name, src_project=self.project.name,
                                dest_model=new_name, dest_project=new_project)
        headers = {'Content-Type': 'application/xml'}
        resp = self._client.post(url, info.serialize(), headers=headers)

        inst_url = resp.headers['Location'].rstrip('/')
        inst_id = urlparse(inst_url).path.rsplit('/', 1)[-1]
        inst = self.project.instances[inst_id]

        if not async_:
            inst.wait_for_success()
            return self.parent[new_name]
        else:
            return inst

    def drop(self):
        self.parent.delete(self)



from libcpp.string cimport string

cdef extern from "<sstream>" namespace "std" nogil:
    cdef cppclass stringstream:
        stringstream() except +
        stringstream(const string &str) except +
        stringstream& push "operator<<" (bint val)
        stringstream& push "operator<<" (short val)
        stringstream& push "operator<<" (unsigned short val)
        stringstream& push "operator<<" (int val)
        stringstream& push "operator<<" (unsigned int val)
        stringstream& push "operator<<" (long val)
        stringstream& push "operator<<" (unsigned long val)
        stringstream& push "operator<<" (float val)
        stringstream& push "operator<<" (double val)
        stringstream& push "operator<<" (long double val)
        stringstream& push "operator<<" (void* val)
        stringstream()
        unsigned int tellp()
        stringstream& put(char c)
        stringstream& write(const char *c, size_t n)
        string to_string "str" () const



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from math import ceil, log, pow


if sys.version < '3':
    PY3 = False
else:
    PY3 = True


def get_SHA1_bin(word):
    """
    Return SHA1 hash of any string

    :param word:
    :return:
    """
    from hashlib import sha1
    if PY3 and isinstance(word, str):
        word = word.encode('utf-8')

    hash_s = sha1()
    hash_s.update(word)
    return bin(int(hash_s.hexdigest(), 16))[2:].zfill(160)


def get_index(binstr, end_index=160):
    """
    Return the position of the first 1 bit
    from the left in the word until end_index

    :param binstr:
    :param end_index:
    :return:
    """
    res = -1
    try:
        res = binstr.index('1') + 1
    except ValueError:
        res = end_index
    return res


class HyperLogLog(object):
    """
    Implements a HyperLogLog
    """

    __ALPHA16 = 0.673
    __ALPHA32 = 0.697
    __ALPHA64 = 0.709

    def __init__(self, error_rate, splitter=None):
        if not (0 < error_rate < 1):
            raise ValueError('error_rate must be between 0 and 1.')

        self._k = int(ceil(log(pow(1.04 / error_rate, 2), 2)))
        self._bucket_number = 1 << self._k
        self._alpha = self.__get_alpha(self._bucket_number)

        self._splitter = splitter

    def buffer(self):
        return [0] * self._bucket_number

    def __get_alpha(self, m):
        if m <= 16:
            return self.__ALPHA16
        elif m <= 32:
            return self.__ALPHA32
        elif m <= 64:
            return self.__ALPHA64
        else:
            return 0.7213 / (1 + 1.079 / float(m))

    def __call__(self, buffer, item):
        """
        Add the items to the HyperLogLog.

        :param buffer:
        :param item:
        :return:
        """

        items = [str(item), ]
        if self._splitter is not None:
            items = str(item).split(self._splitter)

        for item in items:
            binword = get_SHA1_bin(item)
            pos = int(binword[:self._k], 2)
            # retrieve the position of leftmost 1
            aux = get_index(binword[self._k:], 160 - self._k)
            # set its own register value to maximum value seen so far
            buffer[pos] = max(aux, buffer[pos])

    def _estimate(self, buffer):
        """
        Return the estimate of the cardinality

        :return: esitimate of the cardinality
        """
        m = self._bucket_number
        raw_e = self._alpha * pow(m, 2) / sum([pow(2, -x) for x in buffer])
        if raw_e <= 5 / 2.0 * m:
            v = buffer.count(0)
            if v != 0:
                return m * log(m / float(v), 2)
            else:
                return raw_e
        elif raw_e <= 1 / 30.0 * 2 ** 160:
            return raw_e
        else:
            return -2 ** 160 * log(1 - raw_e / 2.0 ** 160, 2)

    def getvalue(self, buffer):
        return int(round(self._estimate(buffer)))

    def merge(self, buffer, other_hyper_log_log):
        """
        Merge the HyperLogLog

        :param other_hyper_log_log:
        :return:
        """

        for i in range(len(buffer)):
            buffer[i] = max(buffer[i], other_hyper_log_log[i])

    def _word_size_calculator(self, n_max):
        """
        Estimate the size of the memory units, using the maximum cardinality as an argument
        :param n_max: maximum cardinality
        :return: size of the memory units
        """

        return int(ceil(log(log(n_max, 2), 2)))



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import textwrap
from collections import OrderedDict
from functools import partial

from .objects import AlgorithmsDef, AlgorithmDef
from ..expr import AlgoCollectionExpr, ODPSModelExpr, MetricsResultExpr
from ..enums import PortType
from ..utils import import_class_member, get_function_args, ML_ARG_PREFIX
from ...compat import six, Iterable
from ...df.expr.collections import SequenceExpr
from ...utils import camel_to_underline, underline_to_capitalized, load_static_text_file,\
    load_internal_static_text_file, survey

SYSTEM_PARAMS = set("""
isInheritLastType inputTableName inputTablePartitions outputTableName outputTablePartition
""".strip().split())
PREDICTOR_PARAMS = set("""
inputTableName inputPartitions modelTableName outputTableName outputPartition appendColNames
""".strip().split())


def dispatch_args(param_names, *args, **kwargs):
    from .base_algo import is_ml_object
    ml_args = [a for a in args if is_ml_object(a)]
    p_args = [a for a in args if not is_ml_object(a)]
    ml_kw = OrderedDict([(k, v) for k, v in six.iteritems(kwargs)
                         if k not in param_names or k.startswith('_')])
    p_kw = OrderedDict([(k, v) for k, v in six.iteritems(kwargs) if k not in ml_kw and k in param_names])
    return ml_args, p_args, ml_kw, p_kw


def _generate_class_ctor(nested_cls, algo_def):
    for param in algo_def.params:
        if not param.value:
            param.value = None
    metas = OrderedDict([(meta.name, meta.value) for meta in algo_def.metas])
    ports = OrderedDict([(port.name, port) for port in algo_def.ports])

    def _ctor(self, *args, **kwargs):
        def _fill_param(param_obj):
            param_obj = param_obj.copy()
            arg_name = param_obj.friendly_name
            if arg_name in kwargs:
                arg_val = kwargs[arg_name]
                if isinstance(arg_val, SequenceExpr):
                    param_obj.value = arg_val.name
                else:
                    param_obj.value = arg_val
            elif param_obj.sequence is not None and param_obj.sequence <= len(args):
                param_obj.value = args[param_obj.sequence - 1]
            return param_obj

        parameters = OrderedDict([(param.name, _fill_param(param)) for param in algo_def.params])

        sequences = set([p.sequence for p in six.itervalues(parameters)])
        for idx, obj in enumerate(args):
            if (idx + 1) not in sequences:
                raise ValueError('Unexpected sequential argument: {0}.'.format(obj))
        names = set([p.friendly_name for p in six.itervalues(parameters)])
        for k, v in six.iteritems(kwargs):
            if k.startswith('_'):  # predefined args
                continue
            if k not in names:
                raise ValueError('Unexpected keyword argument: {0}={1}.'.format(k, v))

        self._reload_fields = algo_def.reload_fields

        super(nested_cls, self).__init__(algo_def.code_name, parameters, ports, metas)

    if metas.get('survey') == 'true':
        return survey(_ctor)
    return _ctor


def _iter_param_cmp_table(params):
    param_pairs = [(camel_to_underline(p.name), p.name) for p in params if p.name not in SYSTEM_PARAMS]
    method_header_len = len(max(param_pairs, key=lambda p: len(p[0]))[0]) + 1
    param_header_len = len(max(param_pairs, key=lambda p: len(p[1]))[1]) + 1

    sep_line = ' ' * 4 + '=' * method_header_len + ' ' + '=' * param_header_len
    yield sep_line
    yield ' ' * 4 + 'SDK' + ' ' * (method_header_len - len('SDK') + 1) + 'PAI Command'
    yield sep_line

    for method_name, param_name in param_pairs:
        yield ' ' * 4 + method_name + ' ' * (method_header_len - len(method_name) + 1) + param_name
    yield sep_line


def _generate_prop(param_name):
    def _getter(self):
        return self._parameters[param_name].value

    def _setter(self, value):
        if isinstance(value, SequenceExpr):
            value = value.name
        self._parameters[param_name].value = value
        self._invoke_setter(param_name, value)

    return property(_getter, _setter)


def _generate_prop_setter(param_name):
    def _setter(self, value):
        if isinstance(value, SequenceExpr):
            value = value.name
        self._parameters[param_name].value = value
        self._invoke_setter(param_name, value)
        return self

    return _setter


def _generate_algo_func(cls, node_def, func_name):
    param_names = set(p.friendly_name for p in node_def.params)

    def _func(*args, **kwargs):
        ml_args, p_args, ml_kw, p_kw = dispatch_args(param_names, *args, **kwargs)
        algo_obj = cls(*p_args, **p_kw)
        return getattr(algo_obj, func_name)(*ml_args, **ml_kw)

    _func.__name__ = ('' if node_def.public else '_') + camel_to_underline(cls.__name__.lstrip('_'))
    _func.__doc__ = _build_function_docstring(node_def)
    return _func


def _is_param_in_doc(param_def):
    pname = param_def.name
    if pname in SYSTEM_PARAMS:
        return False
    if pname.endswith('TableName') or pname.endswith('Partition') or pname.endswith('Partitions'):
        return False
    if pname.endswith('Name') and 'Table' in pname:
        return False
    if pname == 'modelName':
        return False
    return True


def _get_port_type(port_def, req=False, sphinx_wrap=False):
    if port_def.type == PortType.DATA:
        data_type = 'DataFrame'
    elif port_def.type == PortType.MODEL:
        if not port_def.model or not port_def.model.type:
            data_type = 'PmmlModel'
        else:
            data_type = port_def.model.type
    else:
        return None
    if sphinx_wrap:
        data_type = ':class:`%s`' % data_type
    if req and port_def.required:
        return data_type + ', required'
    else:
        return data_type


def _build_node_def_params(node_def, filter_func=None):
    filter_func = filter_func or (lambda p: True)
    param_docs = []
    for p in node_def.params:
        if not filter_func(p):
            continue
        if not _is_param_in_doc(p):
            continue
        if p.internal:
            continue
        if p.docs:
            param_docs.append(six.text_type('    - **{param_name}** - {docs}').format(
                param_name=p.friendly_name, docs=re.sub(r'\n *', ' ', p.docs))
            )
        else:
            param_docs.append('    - **%s**' % p.friendly_name)
    if param_docs:
        return '\n'.join(param_docs) + '\n'
    else:
        return ''


def _filter_predictor_params(p):
    if p.name in PREDICTOR_PARAMS:
        return False
    if p.exporter and 'get_input_model_param' in p.exporter:
        return False
    return True


def _build_class_docstring(node_def_dict, code_name):
    node_def = node_def_dict[code_name]
    if node_def.public is not None and not node_def.public:
        return ''
    docs = six.text_type(node_def.docs) if node_def.docs else six.text_type('\n%params%\n%predictor_params%\n')
    docs = textwrap.dedent(docs)
    docs = docs.replace('%params%', ':Parameters:\n' + _build_node_def_params(node_def))

    predictor_code_name = node_def.meta_dict.get('predictor')
    if predictor_code_name:
        predictor_node = node_def_dict[predictor_code_name]
        predictor_input_docs = []
        for p in predictor_node.get_input_ports():
            if p.name == 'model':
                continue
            if p.docs:
                predictor_input_docs.append(six.text_type('    - **{param_name}** ({param_type}) - {docs}').format(
                    param_type=_get_port_type(p, sphinx_wrap=True),
                    param_name=camel_to_underline(p.name),
                    docs=re.sub(r'\n *', ' ', p.docs),
                ))
            else:
                predictor_input_docs.append(six.text_type('    - **{param_name}** ({param_type})').format(
                    param_type=_get_port_type(p, sphinx_wrap=True),
                    param_name=camel_to_underline(p.name),
                ))
        predictor_docs = ':Predictor Parameters:\n' + \
                         '\n'.join(predictor_input_docs) + '\n' +\
                         _build_node_def_params(predictor_node, _filter_predictor_params)
        docs = docs.replace('%predictor_params%', predictor_docs)
    else:
        docs = docs.replace('%predictor_params%', '')
    return docs


def _build_entry_method_docs(node_def):
    docs = six.text_type(node_def.entry_docs) if node_def.entry_docs else six.text_type('\n%inputs%\n%outputs%\n')
    if '%inputs%' in docs:
        input_docs = [six.text_type(':param %s %s: %s') % (_get_port_type(p), camel_to_underline(p.name), re.sub(r'\n *', ' ', p.docs))
                      if p.docs else six.text_type(':param %s %s:') % (_get_port_type(p), camel_to_underline(p.name))
                      for p in node_def.get_input_ports()]
        if input_docs:
            docs = docs.replace('%inputs%', six.text_type('.. py:currentmodule:: odps.ml\n\n') + '\n'.join(input_docs) + '\n\n')
        else:
            docs = docs.replace('%inputs%', '')
    if '%outputs%' in docs:
        output_ports = node_def.get_output_ports()
        if len(output_ports) == 1:
            if output_ports[0].docs:
                output_docs = '\n\n:return: %s\n:rtype: %s\n\n' % (output_ports[0].docs, _get_port_type(output_ports[0]))
            else:
                output_docs = '\n\n:rtype: %s\n\n' % _get_port_type(output_ports[0])
            docs = docs.replace('%outputs%', output_docs)
        else:
            output_names = [camel_to_underline(p.name) for p in output_ports]
            output_types = [_get_port_type(p) for p in output_ports]
            output_descs = [u'    * %s – %s' % (camel_to_underline(p.name), p.docs) for p in output_ports if p.docs]

            if output_descs:
                output_docs = six.text_type('\n\n:Returns:\n    * (%s)\n%s\n:rtype: (%s)\n\n') % (', '.join(output_names), '\n'.join(output_descs), ', '.join(output_types))
            else:
                output_docs = six.text_type('\n\n:Returns: (%s)\n:rtype: (%s)\n\n') % (', '.join(output_names), ', '.join(output_types))
            docs = docs.replace('%outputs%', output_docs)
    return docs


def _build_entry_method(class_obj, node_def):
    def _entry_method(self, *args, **kwargs):
        return getattr(super(class_obj, self), class_obj._entry_method)(*args, **kwargs)

    _entry_method.__name__ = class_obj._entry_method
    _entry_method.__doc__ = _build_entry_method_docs(node_def)
    _entry_method.__dict__.update(getattr(class_obj, class_obj._entry_method).__dict__)
    return _entry_method


def _build_function_docstring(node_def):
    if node_def.public is not None and not node_def.public:
        return ""
    docs = six.text_type(node_def.docs) if node_def.docs else six.text_type('\n%params%\n')
    docs = textwrap.dedent(docs)
    if '%params%' in docs:
        arg_str = ''
        input_docs = [six.text_type(':param %s %s: %s') % (_get_port_type(p, False, False), camel_to_underline(p.name), re.sub(r'\n *', ' ', p.docs))
                      if p.docs else six.text_type(':param %s %s: %s') % (_get_port_type(p, False, False), camel_to_underline(p.name), camel_to_underline(p.name))
                      for p in node_def.get_input_ports()]
        arg_str += '\n'.join(input_docs) + '\n'
        param_docs = [six.text_type(':param %s: %s') % (p.friendly_name, re.sub(r'\n *', ' ', p.docs))
                      if p.docs else ':param %s: %s' % (p.friendly_name, p.friendly_name)
                      for p in node_def.params if _is_param_in_doc(p) and not p.internal]
        arg_str += '\n'.join(param_docs)
        if arg_str:
            docs = docs.replace('%params%', six.text_type('.. py:currentmodule:: odps.ml\n\n') + arg_str + '\n\n')
        else:
            docs = docs.replace('%params%', '\n\n')
    return docs

_PARAMED_EXPORTER_REGEX = re.compile(r'([^\(]+)\(([^\)]+)\) *')


def _build_expr(node_def, base_class):
    expr_name = underline_to_capitalized(node_def.code_name)
    input_names = tuple([ML_ARG_PREFIX + p.name for p in node_def.get_input_ports()])
    exported = set(p.name for p in node_def.params if p.exported)
    metas = OrderedDict([(meta.name, meta.value) for meta in node_def.metas])

    from ..expr import exporters
    exporter_dict = dict()
    for param in (p for p in node_def.params if p.exporter is not None):
        def fetch_exporter_func(func_name):
            match = _PARAMED_EXPORTER_REGEX.match(func_name)
            arg_dict = dict()
            if match:
                func_name, arg_str = match.group(1), match.group(2)
                arg_str = arg_str.replace(r'\,', '\x01')
                for arg_desc in arg_str.split(','):
                    arg_desc = arg_desc.replace('\x01', ',')
                    k, v = arg_desc.strip().split('=', 1)
                    arg_dict[k.strip()] = v.strip()
            if '.' not in func_name:
                func = getattr(exporters, func_name)
            else:
                func = import_class_member(func_name)
            if not arg_dict:
                return func
            else:
                return partial(func, **arg_dict)

        args = dict()
        exporter_func = fetch_exporter_func(param.exporter)
        for arg in get_function_args(exporter_func):
            if arg == 'param_name':
                args[arg] = param.name
            elif arg == 'param_value':
                args[arg] = param.value
            elif arg == 'input_name':
                args[arg] = param.input_name
            elif arg == 'output_name':
                args[arg] = param.output_name

        def make_exporter_wrapper(exporter_name, args_dict):
            func = fetch_exporter_func(exporter_name)
            exporter_fun_name = exporter_name
            if '.' in exporter_fun_name:
                _, exporter_fun_name = exporter_fun_name.rsplit('.', 1)
            fun = lambda self: func(self, **args_dict)
            fun.__name__ = exporter_fun_name
            return fun

        exporter_dict[param.name] = make_exporter_wrapper(param.exporter, args)

    class_dict = {
        'node_name': node_def.code_name,
        '_algo': node_def.code_name,
        '_args': input_names,
        '_exported': exported,
        '_exporters': exporter_dict,
        'input_ports': node_def.get_input_ports(),
        'output_ports': node_def.get_output_ports(),
        'algo_meta': metas,
    }
    return type(base_class)(expr_name + base_class._suffix, (base_class, ), class_dict)


def _generate_objects(def_list):
    classes = []
    node_def_dict = dict((d.code_name, d) for d in def_list)
    for node_def in (d for d in def_list if d.enabled is None or d.enabled):
        class_name = underline_to_capitalized(node_def.code_name)
        if node_def.public is not None and not node_def.public:
            class_name = '_' + class_name
        base_class = import_class_member('$package_root.algolib.base_algo.' + node_def.base_class)

        if not node_def.export_function:
            doc_str = _build_class_docstring(node_def_dict, node_def.code_name)
        else:
            doc_str = None
        new_cls = type(class_name, (base_class, ), {'__doc__': doc_str})

        setattr(new_cls, '__init__', _generate_class_ctor(new_cls, node_def))
        for param in (p for p in node_def.params if p.name not in SYSTEM_PARAMS):
            user_param_name = param.friendly_name
            setattr(new_cls, user_param_name, _generate_prop(param.name))
            setattr(new_cls, 'set_' + user_param_name, _generate_prop_setter(param.name))

        if new_cls._entry_method:
            setattr(new_cls, new_cls._entry_method, _build_entry_method(new_cls, node_def))

        new_cls._collection_expr = _build_expr(node_def, AlgoCollectionExpr)
        new_cls._model_expr = _build_expr(node_def, ODPSModelExpr)
        if base_class.__name__ == 'BaseMetricsAlgorithm':
            new_cls._metrics_expr = _build_expr(node_def, MetricsResultExpr)

        classes.append(new_cls)

        if node_def.export_function:
            if base_class.__name__ == 'BaseMetricsAlgorithm':
                func_name = 'calc'
            else:
                func_name = 'transform'
            classes.append(_generate_algo_func(new_cls, node_def, func_name))

    return classes


def load_algorithms(json_obj, base_class, env):
    def_list = [def_obj if isinstance(def_obj, AlgorithmDef) else AlgorithmDef.parse(def_obj)
                for def_obj in json_obj]

    for def_obj in def_list:
        if not def_obj.base_class:
            def_obj.base_class = base_class

    objects = sorted(_generate_objects(def_list), key=lambda o: o.__name__.lower())
    for algo_obj in objects:
        algo_obj.__module__ = env.__name__
        setattr(algo_obj, '_env', env.__name__)
        setattr(env, algo_obj.__name__, algo_obj)
    names = [c.__name__ for c in objects]

    # add names into __all__ variable for sphinx docs
    if not hasattr(env, '__all__'):
        setattr(env, '__all__', [])
    existing = set(getattr(env, '__all__')) - set(["Enum", ])
    setattr(env, '__all__', list(existing) + list(filter(lambda v: v not in existing and not v.startswith('_'), names)))


def load_defined_algorithms(env, name):
    algos_all = []
    external_xml = load_static_text_file('algorithms/' + name + '.xml')
    base_class = None
    if external_xml:
        algos = AlgorithmsDef.parse(external_xml)
        base_class = algos.base_class
        for algo in (a for a in algos.algorithms if not a.base_class):
            algo.base_class = algos.base_class
        algos_all.extend(algos.algorithms)

    internal_xml = load_internal_static_text_file('algorithms/' + name + '.xml')
    if internal_xml:
        algos = AlgorithmsDef.parse(internal_xml)
        base_class = algos.base_class
        for algo in (a for a in algos.algorithms if not a.base_class):
            algo.base_class = algos.base_class
        algos_all.extend(algos.algorithms)

    load_algorithms(algos_all, base_class, env)


def load_classifiers(algo_defs, env):
    """
    Load an algorithm into a module. The algorithm is an instance of ``AlgorithmDef`` class.

    :param algo_defs: algorithm definitions
    :type algo_defs: AlgorithmDef | list[AlgorithmDef]
    :param env: environment

    :Example:
    >>> import sys
    >>> from odps.ml.algolib.loader import *
    >>> a = XflowAlgorithmDef('SampleAlgorithm')
    >>> a.add_param(ParamDef('param1', 'val1'))
    >>> a.add_port(PortDef('input'))
    >>> a.add_port(PortDef('output', PortDirection.OUTPUT))
    >>> load_classifiers(a, sys.modules[__name__])
    """
    if not isinstance(algo_defs, Iterable):
        algo_defs = [algo_defs, ]
    load_algorithms(algo_defs, 'BaseTrainingAlgorithm', env)


def load_process_algorithm(algo_defs, env):
    """
    Load an algorithm into a module. The algorithm is an instance of ``AlgorithmDef`` class.

    :param algo_defs: algorithm definitions
    :type algo_defs: AlgorithmDef | list[AlgorithmDef]
    :param env: environment

    :Example:
    >>> import sys
    >>> from odps.ml.algolib.loader import *
    >>> a = XflowAlgorithmDef('SampleAlgorithm')
    >>> a.add_param(ParamDef('param1', 'val1'))
    >>> a.add_port(PortDef('input'))
    >>> a.add_port(PortDef('output', PortDirection.OUTPUT))
    >>> load_process_algorithm(a, sys.modules[__name__])
    """
    if not isinstance(algo_defs, Iterable):
        algo_defs = [algo_defs, ]
    load_algorithms(algo_defs, 'BaseProcessAlgorithm', env)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from odps.compat import six

from ..rpc import SandboxRpcChannel, CupidRpcController
from ..errors import CupidError

try:
    from ..proto import kv_store_service_pb2 as kv_pb
except TypeError:
    warnings.warn('Cannot import protos from pycupid: '
        'consider upgrading your protobuf python package.', ImportWarning)
    raise ImportError


class CupidKVStore(object):
    def __init__(self):
        self.channel = SandboxRpcChannel()
        self.stub = kv_pb.KVStoreService_Stub(self.channel)

    def __setitem__(self, key, value):
        if isinstance(value, six.text_type):
            value = value.encode()

        controller = CupidRpcController()
        req = kv_pb.PutRequest(key=key, value=value)

        self.stub.Put(controller, req, None)
        if controller.Failed():
            raise CupidError(controller.ErrorText())

    def __getitem__(self, item):
        controller = CupidRpcController()

        req = kv_pb.GetRequest(value=item)
        resp = self.stub.Get(controller, req, None)

        if controller.Failed():
            err_text = controller.ErrorText()
            if 'PANGU_FILE_NOT_FOUND' in err_text:
                raise KeyError(err_text)
            else:
                raise CupidError(err_text)
        return resp.value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time

from ..analyzer import BaseAnalyzer
from ...expr.expressions import *
from ...expr.element import IntToDatetime
from ...expr.arithmetic import Power, Equal
from ...expr.datetimes import UnixTimestamp
from ...expr.reduction import GroupedSequenceReduction, \
    Moment, GroupedMoment, Kurtosis, GroupedKurtosis

try:
    import numpy as np
except ImportError:
    pass


class Analyzer(BaseAnalyzer):
    def visit_builtin_function(self, expr):
        try:
            collection = six.next(iter(n for n in self._dag.iter_descendants(expr)
                                       if isinstance(n, CollectionExpr))).input
        except StopIteration:
            raise NotImplementedError

        if isinstance(expr, RandomScalar):
            seed = expr._func_args[0] if len(expr._func_args) >= 1 else None
            if seed is not None:
                np.random.seed(seed)

            col = getattr(collection, collection.schema.names[0])
            self._sub(expr, col.map(lambda v: np.random.rand()).astype('float'))
        else:
            raise NotImplementedError

    def visit_element_op(self, expr):
        if isinstance(expr, IntToDatetime):
            sub = expr.input.map(lambda n: datetime.datetime.fromtimestamp(n))
            self._sub(expr, sub)
        else:
            raise NotImplementedError

    def visit_binary_op(self, expr):
        if isinstance(expr, Power) and isinstance(expr._lhs.dtype, types.Integer) and \
                isinstance(expr._rhs.dtype, types.Integer):
            self._sub(expr._lhs, expr._lhs.astype('float'), parents=[expr])
        else:
            raise NotImplementedError

    def visit_datetime_op(self, expr):
        if isinstance(expr, UnixTimestamp):
            sub = expr.input.map(lambda d: time.mktime(d.timetuple()))
            self._sub(expr, sub)
        else:
            raise NotImplementedError

    def visit_reduction(self, expr):
        if isinstance(expr, (Moment, GroupedMoment)):
            order = expr._order
            center = expr._center

            sub = self._get_moment_sub_expr(expr, expr.input, order, center)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
            return
        elif isinstance(expr, (Kurtosis, GroupedKurtosis)):
            std = expr.input.std(ddof=0)
            if isinstance(expr, GroupedSequenceReduction):
                std = std.to_grouped_reduction(expr._grouped)
            cnt = expr.input.count()
            if isinstance(expr, GroupedSequenceReduction):
                cnt = cnt.to_grouped_reduction(expr._grouped)
            m4 = self._get_moment_sub_expr(expr, expr.input, 4, True)
            sub = 1.0 / (cnt - 2) / (cnt - 3) * ((cnt * cnt - 1) * m4 / (std ** 4) - 3 * (cnt - 1) ** 2)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
            return

        raise NotImplementedError

    def visit_join(self, expr):
        for p in expr._predicate:
            if isinstance(p, BooleanSequenceExpr):
                es = (expr for expr in p.traverse(top_down=True, unique=True,
                                                  stop_cond=lambda x: x is p._lhs or x is p._rhs))
                for e in es:
                    if isinstance(e, Equal):
                        expr._reverse_equal(e)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import OrderedDict

from .core import Iterable, XMLRemoteModel
from .xflow import XFlow
from .instance import Instance
from .. import serializers, errors, compat, options
from ..compat import six


class XFlows(Iterable):

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    xflows = serializers.XMLNodesReferencesField(XFlow, 'odpsalgo')

    def _get(self, name):
        return XFlow(client=self._client, parent=self, name=name)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            xflow = self._get(item)
        elif isinstance(item, XFlow):
            xflow = item
        else:
            return False

        try:
            xflow.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self, owner=None):
        params = dict()
        if owner is not None:
            params['owner'] = owner

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            inst = XFlows.parse(self._client, resp, obj=self)
            params['marker'] = inst.marker

            return inst.xflows

        while True:
            xflows = _it()
            if xflows is None:
                break
            for xflow in xflows:
                yield xflow

    def create(self, xml_source):
        url = self.resource()
        headers = {'Content-Type': 'application/xml'}
        self._client.post(url, xml_source, headers=headers)

    def delete(self, name):
        if not isinstance(name, XFlow):
            xflow = XFlow(name=name, parent=self)
        else:
            xflow = name
            name = name.name
        del self[name]

        url = xflow.resource()
        self._client.delete(url)

    def update(self, xflow):
        url = xflow.resource()
        headers = {'Content-Type': 'application/xml'}
        self._client.put(url, xflow.xml_source, headers)

        xflow.reload()
        return xflow

    class XFlowInstance(XMLRemoteModel):
        __slots__ = 'xflow_project', 'xflow_name', 'parameters', 'priority', 'properties'
        _root = 'XflowInstance'

        xflow_project = serializers.XMLNodeField('Project')
        xflow_name = serializers.XMLNodeField('Xflow')
        parameters = serializers.XMLNodePropertiesField(
            'Parameters', 'Parameter', key_tag='Key', value_tag='Value', required=True
        )
        priority = serializers.XMLNodeField(
            'Priority', parse_callback=int, serialize_callback=int
        )
        properties = serializers.XMLNodePropertiesField(
            'Config', 'Property', key_tag='Name', value_tag='Value'
        )

    class AnonymousSubmitXFlowInstance(XMLRemoteModel):
        _root = 'Instance'

        instance = serializers.XMLNodeReferenceField('XFlows.XFlowInstance', 'XflowInstance')

    @staticmethod
    def _gen_xflow_instance_xml(xflow_instance=None, **kw):
        if xflow_instance is None:
            xflow_instance = XFlows.XFlowInstance(**kw)

        inst = XFlows.AnonymousSubmitXFlowInstance(instance=xflow_instance)
        return inst.serialize()

    def run_xflow(self, xflow_instance=None, project=None, hints=None, parameters=None, **kw):
        project = project or self.parent
        hints = hints or {}
        if options.ml.xflow_settings:
            hints.update(options.ml.xflow_settings)
        if hints:
            kw['properties'] = hints
        if options.biz_id:
            if kw.get('properties') is None:
                kw['properties'] = OrderedDict()
            kw['properties']['biz_id'] = str(options.biz_id)
        if parameters:
            new_params = OrderedDict()
            for k, v in six.iteritems(parameters):
                if k == 'modelName' and '/' not in v:
                    new_params[k] = '%s/offlinemodels/%s' % (project.name, v)
                elif k in ('inputTableName', 'outputTableName') and '.' not in v:
                    new_params[k] = '%s.%s' % (project.name, v)
                else:
                    new_params[k] = v
            parameters = new_params
        return project.instances.create(
            xml=self._gen_xflow_instance_xml(xflow_instance=xflow_instance, parameters=parameters, **kw))

    class XFlowResult(XMLRemoteModel):
        class XFlowAction(XMLRemoteModel):
            node_type = serializers.XMLNodeAttributeField('.', attr='NodeType')
            instance_id = serializers.XMLNodeField('InstanceId')
            name = serializers.XMLNodeField('Name')
            result = serializers.XMLNodeReferenceField(Instance.InstanceResult, 'Result')

        actions = serializers.XMLNodesReferencesField(XFlowAction, 'Actions', 'Action')

    def get_xflow_results(self, instance):
        url = instance.resource()
        resp = self._client.get(url, action='xresult')

        xflow_result = XFlows.XFlowResult.parse(self._client, resp)
        return {action.name: action for action in xflow_result.actions}

    def get_xflow_source(self, instance):
        return self._client.get(instance.resource(), action='xsource').content

    def get_xflow_instance(self, instance):
        content = self.get_xflow_source(instance)
        try:
            inst = XFlows.AnonymousSubmitXFlowInstance.parse(self._client, content)
            return inst.instance
        except compat.ElementTreeParseError as e:
            raise errors.ODPSError(e)

    def get_xflow_sub_instances(self, instance):
        inst_dict = OrderedDict()
        for x_result in filter(lambda xr: xr.node_type != 'Local',
                               six.itervalues(self.get_xflow_results(instance))):
            if x_result.node_type == 'Instance':
                inst_dict[x_result.name] = self.parent.instances[x_result.instance_id]
            elif x_result.node_type == 'SubWorkflow':
                sub_instance = self.parent.instances[x_result.instance_id]
                sub_inst_dict = self.get_xflow_sub_instances(sub_instance)
                inst_dict.update(**sub_inst_dict)
        return inst_dict

    def iter_xflow_sub_instances(self, instance, interval=1, check=False):
        inst_id_set = set()
        while not instance.is_terminated(retry=True):
            sub_tasks_result = self.get_xflow_sub_instances(instance)
            for k, v in six.iteritems(sub_tasks_result):
                if v.id not in inst_id_set:
                    inst_id_set.add(v.id)
                    yield k, v
            try:
                time.sleep(interval)
            except KeyboardInterrupt:
                break
        if check:
            instance.wait_for_success(interval=interval)

    def is_xflow_instance(self, instance):
        try:
            self.get_xflow_instance(instance)
            return True
        except errors.ODPSError:
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from ..inter import enter, setup, teardown, list_rooms
from ..compat import six, StringIO
from .. import types as odps_types
from .. import options, ODPS
from ..utils import replace_sql_parameters, init_progress_ui
from ..df import DataFrame, Scalar, NullScalar, Delay
from ..df.backends.frame import ResultFrame
from ..df.backends.odpssql.types import odps_schema_to_df_schema, odps_type_to_df_type
from ..models import TableSchema
from ..ui.common import html_notify
from ..ui.progress import create_instance_group, reload_instance_status, fetch_instance_group

logger = logging.getLogger(__name__)

try:
    import numpy as np

    np_int_types = map(np.dtype, [np.int_, np.int8, np.int16, np.int32, np.int64])
    np_float_types = map(np.dtype, [np.float_, np.float16, np.float32, np.float64])
    np_to_odps_types = dict([(t, odps_types.bigint) for t in np_int_types] +
                            [(t, odps_types.double) for t in np_float_types])
except ImportError:
    pass

try:
    from IPython.core.magic import Magics, magics_class, line_cell_magic, line_magic
except ImportError:
    # skipped for ci
    ODPSSql = None
    pass
else:
    @magics_class
    class ODPSSql(Magics):

        _odps = None

        def _set_odps(self):
            if self._odps is not None:
                return

            if options.account is not None and options.default_project is not None:
                self._odps = ODPS._from_account(
                    options.account, options.default_project,
                    endpoint=options.endpoint, tunnel_endpoint=options.tunnel.endpoint
                )
            else:
                self._odps = enter().odps

        @line_magic('enter')
        def enter(self, line):
            room = line.strip()
            if room:
                r = enter(room)
                self._odps = r.odps
            else:
                r = enter()
                self._odps = r.odps

            if 'o' not in self.shell.user_ns:
                self.shell.user_ns['o'] = self._odps
                self.shell.user_ns['odps'] = self._odps

            return r

        @line_magic('setup')
        def setup(self, line):
            args = line.strip().split()
            name, args = args[0], args[1:]
            setup(*args, room=name)
            html_notify('Setup succeeded')

        @line_magic('teardown')
        def teardown(self, line):
            name = line.strip()
            teardown(name)
            html_notify('Teardown succeeded')

        @line_magic('list_rooms')
        def list_rooms(self, line):
            return list_rooms()

        @line_magic('stores')
        def list_stores(self, line):
            line = line.strip()

            if line:
                room = enter(line)
            else:
                room = enter()

            return room.display()

        @staticmethod
        def _get_task_percent(task_progress):
            if len(task_progress.stages) > 0:
                all_percent = sum((float(stage.terminated_workers) / stage.total_workers)
                                  for stage in task_progress.stages if stage.total_workers > 0)
                return all_percent / len(task_progress.stages)
            else:
                return 0

        def _to_stdout(cls, msg):
            print(msg)

        @line_magic('set')
        def set_hint(self, line):
            if '=' not in line:
                raise ValueError('Hint for sql is not allowed')

            key, val = line.strip().strip(';').split('=', 1)
            key, val = key.strip(), val.strip()

            settings = options.sql.settings
            if settings is None:
                options.sql.settings = {key: val}
            else:
                options.sql.settings[key] = val

        @line_cell_magic('sql')
        def execute(self, line, cell=''):
            self._set_odps()

            content = line + '\n' + cell
            content = content.strip()

            sql = None
            hints = dict()

            splits = content.split(';')
            for s in splits:
                stripped = s.strip()
                if stripped.lower().startswith('set '):
                    hint = stripped.split(' ', 1)[1]
                    k, v = hint.split('=', 1)
                    k, v = k.strip(), v.strip()
                    hints[k] = v
                elif len(stripped) == 0:
                    continue
                else:
                    if sql is None:
                        sql = s
                    else:
                        sql = '%s;%s' % (sql, s)

            # replace user defined parameters
            sql = replace_sql_parameters(sql, self.shell.user_ns)

            if sql:
                progress_ui = init_progress_ui()
                group_id = create_instance_group('SQL Query')
                progress_ui.add_keys(group_id)

                instance = self._odps.run_sql(sql, hints=hints)
                if logger.getEffectiveLevel() <= logging.INFO:
                    logger.info('Instance ID: %s\n  Log view: %s', instance.id, instance.get_logview_address())
                reload_instance_status(self._odps, group_id, instance.id)
                progress_ui.status('Executing')

                percent = 0
                while not instance.is_terminated(retry=True):
                    last_percent = percent

                    reload_instance_status(self._odps, group_id, instance.id)
                    inst_progress = fetch_instance_group(group_id).instances.get(instance.id)

                    if inst_progress is not None and len(inst_progress.tasks) > 0:
                        percent = sum(self._get_task_percent(task)
                                      for task in six.itervalues(inst_progress.tasks)) / len(inst_progress.tasks)
                    else:
                        percent = 0

                    percent = min(1, max(percent, last_percent))

                    progress_ui.update(percent)
                    progress_ui.update_group()

                    time.sleep(1)

                instance.wait_for_success()
                progress_ui.update(1)

                try:
                    with instance.open_reader() as reader:
                        try:
                            import pandas as pd
                            try:
                                from pandas.io.parsers import ParserError as CParserError
                            except ImportError:
                                pass
                            try:
                                from pandas.parser import CParserError
                            except ImportError:
                                CParserError = ValueError

                            if not hasattr(reader, 'raw'):
                                res = ResultFrame([rec.values for rec in reader],
                                                  schema=odps_schema_to_df_schema(reader._schema))
                            else:
                                try:
                                    res = pd.read_csv(StringIO(reader.raw))
                                    if len(res.values) > 0:
                                        schema = DataFrame(res).schema
                                    else:
                                        cols = res.columns.tolist()
                                        schema = odps_schema_to_df_schema(
                                           TableSchema.from_lists(cols, ['string' for _ in cols]))
                                    res = ResultFrame(res.values, schema=schema)
                                except (ValueError, CParserError):
                                    res = reader.raw
                        except ImportError:
                            if not hasattr(reader, 'raw'):
                                res = ResultFrame([rec.values for rec in reader],
                                                  schema=odps_schema_to_df_schema(reader._schema))
                            else:
                                try:
                                    columns = [odps_types.Column(name=col.name,
                                                                 typo=odps_type_to_df_type(col.type))
                                               for col in reader._columns]
                                    res = ResultFrame(list(reader), columns=columns)
                                except TypeError:
                                    res = reader.raw

                    html_notify('SQL execution succeeded')
                    return res
                finally:
                    progress_ui.close()

        @line_magic('persist')
        def persist(self, line):
            try:
                import pandas as pd
                has_pandas = True
            except ImportError:
                has_pandas = False

            self._set_odps()

            line = line.strip().strip(';')

            frame_name, table_name = line.split(None, 1)

            if '.' in table_name:
                parts = table_name.split('.')
                if len(parts) == 3:
                    project_name, schema_name, table_name = parts
                else:
                    project_name, table_name = parts
                    schema_name = None
                if table_name.startswith('`') and table_name.endswith('`'):
                    table_name = table_name[1:-1]
            else:
                project_name = schema_name = None

            frame = self.shell.user_ns[frame_name]
            if self._odps.exist_table(table_name, project=project_name, schema=schema_name):
                raise TypeError('%s already exists' % table_name)

            if isinstance(frame, DataFrame):
                frame.persist(name=table_name, project=project_name, schema=schema_name, notify=False)
            elif has_pandas and isinstance(frame, pd.DataFrame):
                frame = DataFrame(frame)
                frame.persist(name=table_name, project=project_name, schema=schema_name, notify=False)
            html_notify('Persist succeeded')


    def load_ipython_extension(ipython):
        ipython.register_magics(ODPSSql)

        # Do global import when load extension
        ipython.user_ns['DataFrame'] = DataFrame
        ipython.user_ns['Scalar'] = Scalar
        ipython.user_ns['NullScalar'] = NullScalar
        ipython.user_ns['options'] = options
        ipython.user_ns['TableSchema'] = TableSchema
        ipython.user_ns['Delay'] = Delay



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport cython
import sys
import threading
import time
from cpython.datetime cimport (
    PyDateTime_DateTime,
    datetime,
    datetime_year,
    datetime_month,
    datetime_day,
    datetime_hour,
    datetime_minute,
    datetime_second,
    datetime_microsecond,
    datetime_new,
    timedelta_new,
    import_datetime,
)
from datetime import datetime
from libc.stdint cimport int64_t
from libc.time cimport time_t, tm, mktime, localtime, gmtime

try:
    import zoneinfo
except ImportError:
    zoneinfo = None
try:
    import pytz
except ImportError:
    pytz = None

from ..compat import utc
from ..config import options

cdef extern from "timegm.c":
    time_t timegm(tm* t) nogil

cdef bint _is_windows = sys.platform.lower().startswith("win")
cdef bint _is_py3 = sys.version_info[0] == 3
cdef bint _datetime_inited = False
cdef object _time_init_lock = threading.Lock()

import_datetime()

cdef int64_t _antique_mills, _min_datetime_mills
cdef object _py_local_epoch

try:
    _py_local_epoch = datetime.fromtimestamp(0)
except OSError:  # workaround for bpo-29097  # pragma: no cover
    _py_local_epoch = datetime(*time.localtime(0)[:6])

try:
    _antique_mills = time.mktime(
        datetime(1928, 1, 1).timetuple()
    ) * 1000
except OverflowError:
    _antique_mills = int(
        (datetime(1928, 1, 1) - datetime.utcfromtimestamp(0)).total_seconds()
    ) * 1000
_min_datetime_mills = int(
    (datetime.min - datetime.utcfromtimestamp(0)).total_seconds() * 1000
)
_antique_errmsg = 'Date older than 1928/01/01 and may contain errors. ' \
                  'Ignore this error by configuring `options.allow_antique_date` to True.'
_min_datetime_errmsg = 'Date exceed range Python can handle. If you are reading data with tunnel, read '\
                       'the value as None by setting options.tunnel.overflow_date_as_none to True, ' \
                       'or convert the value into strings with SQL before processing them with Python.'

cdef inline bint datetime_hastzinfo(object o):
    return (<PyDateTime_DateTime*>o).hastzinfo


cdef class CMillisecondsConverter:
    @staticmethod
    def _get_tz(tz):
        if type(tz) is unicode or type(tz) is bytes:
            if pytz is None and zoneinfo is None:
                raise ImportError('Package `pytz` is needed when specifying string-format time zone.')
            else:
                return zoneinfo.ZoneInfo(tz) if zoneinfo is not None else pytz.timezone(tz)
        else:
            return tz

    def __init__(self, local_tz=None, is_dst=False):
        global _datetime_inited

        if not _is_py3 and not _datetime_inited:
            # due to cython implementations, make sure the line below is always invoked
            with _time_init_lock:
                if not _datetime_inited:
                    import_datetime()
                    _datetime_inited = True

        self._local_tz = local_tz if local_tz is not None else options.local_timezone
        if self._local_tz is None:
            self._local_tz = True
        self._use_default_tz = type(self._local_tz) is bool
        self._default_tz_local = self._use_default_tz and self._local_tz

        self._allow_antique = options.allow_antique_date or _antique_mills is None
        self._is_dst = is_dst

        self._tz = self._get_tz(self._local_tz) if not self._use_default_tz else None
        self._tz_has_localize = hasattr(self._tz, 'localize')

    cdef int _build_tm_struct(self, datetime dt, tm *p_tm) except? -1:
        p_tm.tm_year = datetime_year(dt) - 1900
        p_tm.tm_mon = datetime_month(dt) - 1
        p_tm.tm_mday = datetime_day(dt)
        p_tm.tm_yday = 0
        p_tm.tm_hour = datetime_hour(dt)
        p_tm.tm_min = datetime_minute(dt)
        p_tm.tm_sec = datetime_second(dt)
        p_tm.tm_isdst = -1

    @cython.cdivision(True)
    cpdef int64_t to_milliseconds(self, datetime dt) except? -1:
        cdef int64_t mills
        cdef tm tm_result
        cdef time_t unix_ts
        cdef bint with_tzinfo = datetime_hastzinfo(dt)
        cdef bint assume_local_tz = self._default_tz_local

        if not self._use_default_tz and not with_tzinfo:
            if self._tz_has_localize:
                dt = self._tz.localize(dt, is_dst=self._is_dst)
            else:
                dt = dt.replace(tzinfo=self._tz)
            with_tzinfo = True

        if with_tzinfo:
            dt = dt.astimezone(utc)
            assume_local_tz = False

        self._build_tm_struct(dt, &tm_result)
        if assume_local_tz:
            if not _is_windows or tm_result.tm_year > 1970:
                unix_ts = mktime(&tm_result)
            else:
                py_datetime = datetime_new(
                    tm_result.tm_year + 1900,
                    tm_result.tm_mon + 1,
                    tm_result.tm_mday,
                    tm_result.tm_hour,
                    tm_result.tm_min,
                    tm_result.tm_sec,
                    0,
                    None,
                )
                unix_ts = int((py_datetime - _py_local_epoch).total_seconds())
        else:
            unix_ts = timegm(&tm_result)

        mills = unix_ts * 1000 + datetime_microsecond(dt) / 1000

        if not self._allow_antique and mills < _antique_mills:
            from ..errors import DatetimeOverflowError

            raise DatetimeOverflowError(_antique_errmsg)
        return mills

    @cython.cdivision(True)
    cpdef datetime from_milliseconds(self, int64_t milliseconds):
        cdef time_t seconds, zero
        cdef int64_t microseconds
        cdef tm* p_tm

        if not self._allow_antique and milliseconds < _antique_mills:
            from ..errors import DatetimeOverflowError

            raise DatetimeOverflowError(_antique_errmsg)
        if milliseconds < _min_datetime_mills:
            from ..errors import DatetimeOverflowError

            raise DatetimeOverflowError(_min_datetime_errmsg)

        if milliseconds >= 0:
            seconds = milliseconds / 1000
            microseconds = milliseconds % 1000 * 1000
        else:
            seconds = milliseconds / 1000
            microseconds = milliseconds % 1000 * 1000
            if microseconds < 0:
                microseconds += 1000000
                seconds -= 1

        if self._use_default_tz:
            if self._default_tz_local and _is_windows and milliseconds < 0:
                # special logic for negative timestamp under Windows
                zero = 0
                p_tm = localtime(&zero)
                return datetime_new(
                    p_tm.tm_year + 1900,
                    p_tm.tm_mon + 1,
                    p_tm.tm_mday,
                    p_tm.tm_hour,
                    p_tm.tm_min,
                    p_tm.tm_sec,
                    microseconds,
                    None,
                ) + timedelta_new(0, seconds, 0)

            if self._default_tz_local:
                p_tm = localtime(&seconds)
            else:
                p_tm = gmtime(&seconds)

            return datetime_new(
                p_tm.tm_year + 1900,
                p_tm.tm_mon + 1,
                p_tm.tm_mday,
                p_tm.tm_hour,
                p_tm.tm_min,
                p_tm.tm_sec,
                microseconds,
                None,
            )
        else:
            return datetime.utcfromtimestamp(seconds) \
                .replace(microsecond=microseconds, tzinfo=utc) \
                .astimezone(self._tz)



# Copyright 1999-2024 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import logging
import multiprocessing
import os
import socket
import struct
import sys
import threading
from types import GeneratorType, MethodType

try:
    import pyarrow as pa
except (AttributeError, ImportError):
    pa = None

from .. import types as odps_types, utils
from ..compat import six
from ..config import options
from ..lib import cloudpickle
from ..lib.tblib import pickling_support
from .readers import TunnelArrowReader, TunnelRecordReader

logger = logging.getLogger(__name__)
pickling_support.install()

_GET_NEXT_BLOCK_CMD = 0x01
_PUT_WRITTEN_BLOCKS_CMD = 0x02
_SERVER_ERROR_CMD = 0xFD
_STOP_SERVER_CMD = 0xFE


if sys.version_info[0] == 2:
    _ord_if_possible = ord

    def _load_classmethod(cls, func_name):
        return getattr(cls, func_name)

    class InstanceMethodWrapper(object):
        """Trick for classmethods under Python 2.7 to be pickleable"""
        def __init__(self, func):
            assert isinstance(func, MethodType)
            assert isinstance(func.im_self, type)
            self._func = func

        def __call__(self, *args, **kw):
            return self._func()

        def __reduce__(self):
            return _load_classmethod, (self._func.im_self, self._func.im_func.__name__)

    _wrap_classmethod = InstanceMethodWrapper
else:
    def _ord_if_possible(x):
        return x

    def _wrap_classmethod(x):
        return x


class SpawnedTableReaderMixin(object):
    @property
    def schema(self):
        return self._parent.table_schema

    @staticmethod
    def _read_table_split(
        conn, download_id, start, count, idx,
        rest_client=None, project=None, table_name=None, partition_spec=None,
        tunnel_endpoint=None, quota_name=None, columns=None, arrow=False,
        schema_name=None
    ):
        # read part data
        from ..tunnel import TableTunnel

        try:
            tunnel = TableTunnel(
                client=rest_client,
                project=project,
                endpoint=tunnel_endpoint,
                quota_name=quota_name,
            )
            session = utils.call_with_retry(
                tunnel.create_download_session,
                table_name,
                schema=schema_name,
                download_id=download_id,
                partition_spec=partition_spec,
            )
            if not arrow:
                data = utils.call_with_retry(
                    session.open_record_reader, start, count, columns=columns
                ).to_pandas()
            else:
                data = utils.call_with_retry(
                    session.open_arrow_reader, start, count, columns=columns
                ).to_pandas()
            conn.send((idx, data, True))
        except:
            conn.send((idx, sys.exc_info(), False))

    def _get_process_split_reader(self, columns=None):
        rest_client = self._parent._client
        table_name = self._parent.name
        schema_name = self._parent.get_schema()
        project = self._parent.project.name
        tunnel_endpoint = self._download_session._client.endpoint
        quota_name = self._download_session._quota_name
        partition_spec = self._partition_spec

        return functools.partial(
            self._read_table_split,
            rest_client=rest_client,
            project=project,
            table_name=table_name,
            partition_spec=partition_spec,
            tunnel_endpoint=tunnel_endpoint,
            quota_name=quota_name,
            arrow=isinstance(self, TunnelArrowReader),
            columns=columns or self._column_names,
            schema_name=schema_name,
        )


class TableRecordReader(SpawnedTableReaderMixin, TunnelRecordReader):
    def __init__(self, table, download_session, partition_spec=None, columns=None):
        super(TableRecordReader, self).__init__(
            table, download_session, columns=columns
        )
        self._partition_spec = partition_spec


class TableArrowReader(SpawnedTableReaderMixin, TunnelArrowReader):
    def __init__(self, table, download_session, partition_spec=None, columns=None):
        super(TableArrowReader, self).__init__(
            table, download_session, columns=columns
        )
        self._partition_spec = partition_spec


class MPBlockServer(object):
    def __init__(self, writer):
        self._writer = writer
        self._sock = None
        self._serve_thread_obj = None
        self._authkey = multiprocessing.current_process().authkey

    @property
    def address(self):
        return self._sock.getsockname() if self._sock else None

    @property
    def authkey(self):
        return self._authkey

    def _serve_thread(self):
        while True:
            data, addr = self._sock.recvfrom(4096)
            try:
                pos = len(self._authkey)
                assert data[:pos] == self._authkey, "Authentication key mismatched!"

                cmd_code = _ord_if_possible(data[pos])
                pos += 1
                if cmd_code == _GET_NEXT_BLOCK_CMD:
                    block_id = self._writer._gen_next_block_id()
                    data = struct.pack("<B", _GET_NEXT_BLOCK_CMD) + struct.pack(
                        "<I", block_id
                    )
                    self._sock.sendto(data, addr)
                elif cmd_code == _PUT_WRITTEN_BLOCKS_CMD:
                    blocks_queue = self._writer._used_block_id_queue
                    count, = struct.unpack("<H", data[pos:pos + 2])
                    pos += 2
                    assert 4 * count < len(data), "Data too short for block count!"
                    block_ids = struct.unpack("<%dI" % count, data[pos:pos + 4 * count])
                    blocks_queue.put(block_ids)
                    self._sock.sendto(struct.pack("<B", _PUT_WRITTEN_BLOCKS_CMD), addr)
                elif cmd_code == _STOP_SERVER_CMD:
                    assert addr[0] == self._sock.getsockname()[0], "Cannot stop server from other hosts!"
                    break
                else:  # pragma: no cover
                    raise AssertionError("Unrecognized command %x", cmd_code)
            except BaseException:
                pk_exc_info = cloudpickle.dumps(sys.exc_info())
                data = struct.pack("<B", _SERVER_ERROR_CMD) + struct.pack(
                    "<I", len(pk_exc_info)
                ) + pk_exc_info
                self._sock.sendto(data, addr)
                logger.exception("Serve thread error.")

        self._sock.close()
        self._sock = None

    def start(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.bind(("127.0.0.1", 0))
        self._serve_thread_obj = threading.Thread(target=self._serve_thread)
        self._serve_thread_obj.daemon = True
        self._serve_thread_obj.start()

    def stop(self):
        stop_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        stop_data = self._authkey + struct.pack("<B", _STOP_SERVER_CMD)
        stop_sock.sendto(stop_data, self._sock.getsockname())
        stop_sock.close()
        self._serve_thread_obj.join()


class MPBlockClient(object):
    _MAX_BLOCK_COUNT = 256

    def __init__(self, address, authkey):
        self._addr = address
        self._authkey = authkey
        self._sock = None

    def __del__(self):
        self.close()

    def close(self):
        if self._sock is not None:
            self._sock.close()
            self._sock = None

    def _get_socket(self):
        if self._sock is None:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return self._sock

    @staticmethod
    def _reraise_remote_error(recv_data):
        if recv_data[0] != _SERVER_ERROR_CMD:
            return
        pk_len, = struct.unpack("<I", recv_data[1:5])
        exc_info = cloudpickle.loads(recv_data[5:5 + pk_len])
        six.reraise(*exc_info)

    def get_next_block_id(self):
        sock = self._get_socket()
        data = self._authkey + struct.pack("<B", _GET_NEXT_BLOCK_CMD)
        sock.sendto(data, self._addr)
        recv_data, server_addr = sock.recvfrom(1024)
        self._reraise_remote_error(recv_data)
        assert _ord_if_possible(recv_data[0]) == _GET_NEXT_BLOCK_CMD
        assert self._addr == server_addr
        count, = struct.unpack("<I", recv_data[1:5])
        return count

    def put_written_blocks(self, block_ids):
        sock = self._get_socket()
        for pos in range(0, len(block_ids), self._MAX_BLOCK_COUNT):
            sub_block_ids = block_ids[pos:pos + self._MAX_BLOCK_COUNT]
            data = (
                self._authkey
                + struct.pack("<B", _PUT_WRITTEN_BLOCKS_CMD)
                + struct.pack("<H", len(sub_block_ids))
                + struct.pack("<%dI" % len(sub_block_ids), *sub_block_ids)
            )
            sock.sendto(data, self._addr)
            recv_data, server_addr = sock.recvfrom(1024)
            self._reraise_remote_error(recv_data)
            assert _ord_if_possible(recv_data[0]) == _PUT_WRITTEN_BLOCKS_CMD
            assert self._addr == server_addr


class AbstractTableWriter(object):
    def __init__(
        self,
        table,
        upload_session,
        blocks=None,
        commit=True,
        on_close=None,
        **kwargs
    ):
        self._table = table
        self._upload_session = upload_session
        self._commit = commit
        self._closed = False
        self._on_close = on_close

        self._use_buffered_writer = None
        if blocks is not None:
            self._use_buffered_writer = False

        # block writer options
        self._blocks = blocks or upload_session.blocks or [0, ]
        self._blocks_writes = [False] * len(self._blocks)
        self._blocks_writers = [None] * len(self._blocks)

        for block in (upload_session.blocks or ()):
            self._blocks_writes[self._blocks.index(block)] = True

        # buffered writer options
        self._thread_to_buffered_writers = dict()

        # objects for cross-process sharings
        self._mp_server = None
        self._main_pid = kwargs.get("main_pid") or os.getpid()
        self._mp_fixed = kwargs.get("mp_fixed")
        if kwargs.get("mp_client"):
            self._mp_client = kwargs["mp_client"]
            self._mp_context = self._block_id_counter = self._used_block_id_queue = None
        else:
            self._mp_client = self._mp_authkey = None
            self._mp_context = kwargs.get("mp_context") or multiprocessing
            self._block_id_counter = kwargs.get(
                "block_id_counter"
            ) or self._mp_context.Value("i", 1 + max(upload_session.blocks or [0]))
            self._used_block_id_queue = kwargs.get(
                "used_block_id_queue"
            ) or self._mp_context.Queue()

    @classmethod
    def _restore_subprocess_writer(
        cls, mp_server_address, mp_server_auth, upload_id, main_pid=None, blocks=None,
        rest_client=None, project=None, table_name=None, partition_spec=None,
        tunnel_endpoint=None, quota_name=None, schema=None
    ):
        from ..core import ODPS
        from ..tunnel import TableTunnel

        odps_entry = ODPS(
            account=rest_client.account,
            app_account=rest_client.app_account,
            endpoint=rest_client.endpoint,
            overwrite_global=False,
        )
        tunnel = TableTunnel(
            client=rest_client,
            project=project,
            endpoint=tunnel_endpoint,
            quota_name=quota_name,
        )
        table = odps_entry.get_table(table_name, schema=schema, project=project)
        session = utils.call_with_retry(
            tunnel.create_upload_session,
            table_name,
            schema=schema,
            upload_id=upload_id,
            partition_spec=partition_spec,
        )
        mp_client = MPBlockClient(mp_server_address, mp_server_auth)
        writer = cls(
            table, session, commit=False, blocks=blocks, main_pid=main_pid,
            mp_fixed=True, mp_client=mp_client,
        )
        return writer

    def _start_mp_server(self):
        if self._mp_server is not None:
            return
        self._mp_server = MPBlockServer(self)
        self._mp_server.start()
        # replace mp queue with ordinary queue
        self._used_block_id_queue = six.moves.queue.Queue()

    def __reduce__(self):
        rest_client = self._table._client
        table_name = self._table.name
        schema_name = self._table.get_schema()
        project = self._table.project.name
        tunnel_endpoint = self._upload_session._client.endpoint
        quota_name = self._upload_session._quota_name
        partition_spec = self._upload_session._partition_spec
        blocks = None if self._use_buffered_writer is not False else self._blocks

        self._start_mp_server()

        return _wrap_classmethod(self._restore_subprocess_writer), (
            self._mp_server.address,
            bytes(self._mp_server.authkey),
            self.upload_id,
            self._main_pid,
            blocks,
            rest_client,
            project,
            table_name,
            partition_spec,
            tunnel_endpoint,
            quota_name,
            schema_name,
        )

    @property
    def upload_id(self):
        return self._upload_session.id

    @property
    def schema(self):
        return self._table.table_schema

    @property
    def status(self):
        return self._upload_session.status

    def _open_writer(self, block_id, compress):
        raise NotImplementedError

    def _write_contents(self, writer, *args):
        raise NotImplementedError

    def _gen_next_block_id(self):
        if self._mp_client is not None:
            return self._mp_client.get_next_block_id()
        with self._block_id_counter.get_lock():
            block_id = self._block_id_counter.value
            self._block_id_counter.value += 1
            return block_id

    def _fix_mp_attributes(self):
        if not self._mp_fixed:
            self._mp_fixed = True
            if os.getpid() == self._main_pid:
                return
            self._commit = False
            self._on_close = None

    def write(self, *args, **kwargs):
        if self._closed:
            raise IOError('Cannot write to a closed writer.')
        self._fix_mp_attributes()

        compress = kwargs.get('compress', False)

        block_id = kwargs.get('block_id')
        if block_id is None:
            if type(args[0]) in six.integer_types:
                block_id = args[0]
                args = args[1:]
            else:
                block_id = 0 if options.tunnel.use_block_writer_by_default else None

        use_buffered_writer = block_id is None
        if self._use_buffered_writer is None:
            self._use_buffered_writer = use_buffered_writer
        elif self._use_buffered_writer is not use_buffered_writer:
            raise ValueError(
                "Cannot mix block writing mode with non-block writing mode within a single writer"
            )

        if use_buffered_writer:
            idx = None
            writer = self._thread_to_buffered_writers.get(threading.current_thread().ident)
        else:
            idx = self._blocks.index(block_id)
            writer = self._blocks_writers[idx]

        if writer is None:
            writer = self._open_writer(block_id, compress)

        self._write_contents(writer, *args)
        if not use_buffered_writer:
            self._blocks_writes[idx] = True

    def close(self):
        if self._closed:
            return

        written_blocks = []
        if self._use_buffered_writer:
            for writer in self._thread_to_buffered_writers.values():
                writer.close()
                written_blocks.extend(writer.get_blocks_written())
        else:
            for writer in self._blocks_writers:
                if writer is not None:
                    writer.close()
            written_blocks = [
                block
                for block, block_write in zip(self._blocks, self._blocks_writes)
                if block_write
            ]

        if written_blocks:
            if self._mp_client is not None:
                self._mp_client.put_written_blocks(written_blocks)
            else:
                self._used_block_id_queue.put(written_blocks)

        if self._commit:
            collected_blocks = []
            # as queue.empty() not reliable, we need to fill local blocks manually
            collected_blocks.extend(written_blocks)
            while not self._used_block_id_queue.empty():
                collected_blocks.extend(self._used_block_id_queue.get())
            collected_blocks.extend(self._upload_session.blocks or [])
            collected_blocks = sorted(set(collected_blocks))
            self._upload_session.commit(collected_blocks)

        if callable(self._on_close):
            self._on_close()

        if self._mp_client is not None:
            self._mp_client.close()
            self._mp_client = None
        if self._mp_server is not None:
            self._mp_server.stop()
            self._mp_server = None
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if an error occurs inside the with block, we do not commit
        if exc_val is not None:
            return
        self.close()


class TableRecordWriter(AbstractTableWriter):
    def _open_writer(self, block_id, compress):
        if self._use_buffered_writer:
            writer = self._upload_session.open_record_writer(
                compress=compress,
                initial_block_id=self._gen_next_block_id(),
                block_id_gen=self._gen_next_block_id,
            )
            thread_ident = threading.current_thread().ident
            self._thread_to_buffered_writers[thread_ident] = writer
        else:
            writer = self._upload_session.open_record_writer(
                block_id, compress=compress
            )
            self._blocks_writers[block_id] = writer
        return writer

    def _to_records(self, arg, sample_rec):
        if odps_types.is_record(sample_rec):
            return arg
        elif isinstance(sample_rec, (list, tuple)):
            return (self._table.new_record(vals) for vals in arg)
        else:
            return None

    def _write_contents(self, writer, *args):
        if len(args) == 0:
            return
        if len(args) > 1:
            args = [args]

        arg = args[0]
        if odps_types.is_record(arg):
            records = [arg]
        elif isinstance(arg, (list, tuple)):
            records = self._to_records(arg, arg[0])
            if records is None:
                records = [self._table.new_record(arg), ]
        elif isinstance(arg, GeneratorType):
            try:
                # peek the first element and then put back
                next_arg = six.next(arg)
                chained = itertools.chain((next_arg,), arg)
                records = self._to_records(chained, next_arg)
            except StopIteration:
                records = ()
        else:
            raise ValueError('Unsupported record type.')

        for record in records:
            writer.write(record)


class TableArrowWriter(AbstractTableWriter):
    def _open_writer(self, block_id, compress):
        if self._use_buffered_writer:
            writer = self._upload_session.open_arrow_writer(
                compress=compress,
                initial_block_id=self._gen_next_block_id(),
                block_id_gen=self._gen_next_block_id,
            )
            thread_ident = threading.current_thread().ident
            self._thread_to_buffered_writers[thread_ident] = writer
        else:
            writer = self._upload_session.open_arrow_writer(
                block_id, compress=compress
            )
            self._blocks_writers[block_id] = writer
        return writer

    def _write_contents(self, writer, *args):
        for arg in args:
            writer.write(arg)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import threading
import multiprocessing
import logging

from odps import options
from odps.compat import six, BytesIO
from odps.tunnel.io.reader import TunnelRecordReader
from odps.tunnel.io.writer import BaseRecordWriter

from cupid.errors import SubprocessStreamEOFError

logger = logging.getLogger(__name__)


class CupidRecordReader(TunnelRecordReader):
    def __init__(self, schema, input_stream, columns=None):
        super(CupidRecordReader, self).__init__(schema, input_stream, columns=columns)
        self._stream = input_stream
        self._table_schema = schema
        self._input_columns = columns
        self._mp_stream = None

    def to_forkable(self):
        return type(self)(self._table_schema, self._create_mp_stream(), columns=self._input_columns)

    def _create_mp_stream(self):
        if isinstance(self._stream, CupidMPInputStream):
            return self._stream
        if self._mp_stream is not None:
            return self._mp_stream

        from multiprocessing.sharedctypes import RawArray
        req_queue = multiprocessing.Queue()
        rep_queue = multiprocessing.Queue()
        buf = RawArray(ctypes.c_char, options.cupid.mp_buffer_size)

        def _mp_thread():
            try:
                while True:
                    read_size = req_queue.get()
                    if read_size < 0:
                        return
                    try:
                        read_size = self._stream.readinto(buf)
                    except SubprocessStreamEOFError:
                        rep_queue.put(-1)
                        return
                    rep_queue.put(read_size)
            finally:
                self.close()

        stream_thread = threading.Thread(target=_mp_thread)
        stream_thread.daemon = True
        stream_thread.start()
        self._mp_stream = CupidMPInputStream(buf, req_queue, rep_queue)
        return self._mp_stream

    def read(self):
        try:
            return super(CupidRecordReader, self).read()
        except SubprocessStreamEOFError:
            return None

    def close(self):
        super(CupidRecordReader, self).close()
        self._stream.close()


class CupidMPInputStream(object):
    def __init__(self, share_buf=None, req_queue=None, rep_queue=None):
        self._buf = share_buf
        self._buf_len = 0
        self._pos = 0
        self._q_req = req_queue
        self._q_rep = rep_queue
        self._eof = False

    def read(self, size):
        if self._eof:
            raise SubprocessStreamEOFError
        if self._pos + size <= self._buf_len:
            buf = bytes(self._buf[self._pos:self._pos + size])
            self._pos += size
            return buf
        else:
            size_left = size
            bio = BytesIO()
            while size_left:
                c = self._buf[self._pos:min(self._pos + size_left, self._buf_len)]
                bio.write(c)
                size_left -= len(c)
                self._pos += len(c)
                if size_left <= 0:
                    break
                self._q_req.put(options.cupid.mp_buffer_size)
                recv_size = self._q_rep.get()
                if recv_size < 0:
                    self._eof = True
                    break
                self._buf_len = recv_size
                self._pos = 0
            return bio.getvalue()

    def close(self):
        self._q_req.put(-1)


class CupidRecordWriter(BaseRecordWriter):
    def __init__(self, schema, output_stream, encoding='utf-8'):
        super(CupidRecordWriter, self).__init__(schema, output_stream, encoding=encoding)
        self._stream = output_stream
        self._block_id = None
        self._partition_spec = None
        self._table_schema = schema
        self._mp_stream = None

    def __getstate__(self):
        return self._table_schema, self._create_mp_stream(), self._block_id, self._partition_spec

    def __setstate__(self, state):
        self.__init__(*state[:2])
        self._block_id, self._partition_spec = state[2:]

    def _create_mp_stream(self):
        if isinstance(self._stream, CupidMPInputStream):
            return self._mp_stream
        if self._mp_stream is not None:
            return self._mp_stream

        from multiprocessing.sharedctypes import RawArray
        req_queue = multiprocessing.Queue()
        rep_queue = multiprocessing.Queue()
        buf = RawArray(ctypes.c_char, options.cupid.mp_buffer_size)

        def _mp_thread():
            try:
                while True:
                    size = req_queue.get()
                    if size < 0:
                        break
                    self.write(buf[:size])
                    rep_queue.put(size)
            finally:
                self.close()

        stream_thread = threading.Thread(target=_mp_thread)
        stream_thread.daemon = True
        stream_thread.start()
        self._mp_stream = CupidMPOutputStream(buf, req_queue, rep_queue)
        return self._mp_stream

    @property
    def block_id(self):
        return self._block_id

    @property
    def partition_spec(self):
        return self._partition_spec

    def close(self):
        super(CupidRecordWriter, self).close()

        # sync by get result
        result = self._stream.result()
        logger.debug('Result fetched on writer close: %s', result)

        self._stream.close()


class CupidMPOutputStream(object):
    def __init__(self, share_buf=None, req_queue=None, rep_queue=None):
        self._buf = share_buf
        self._buf_len = len(share_buf)
        self._pos = 0
        self._q_req = req_queue
        self._q_rep = rep_queue

    def write(self, data):
        data_pos = 0
        data_len = len(data)
        while data_pos < data_len:
            buffer_bound = min(self._buf_len, self._pos + data_len - data_pos)
            data_bound = data_pos + buffer_bound - self._pos
            self._buf[self._pos:buffer_bound] = data[data_pos:data_bound]
            data_pos = data_bound
            self._pos = buffer_bound

            if self._pos == self._buf_len:
                self.flush()
        self.flush()

    def flush(self):
        if self._pos == 0:
            return
        self._q_req.put(self._pos)
        self._q_rep.get()
        self._pos = 0

    def close(self):
        self._q_req.put(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from .core import Backend
from ..utils import traverse_until_source
from ..expr.expressions import Scalar, SequenceExpr, CollectionExpr
from ..expr.reduction import GroupedSequenceReduction
from ..expr.element import Switch
from .. import output
from ... import compat
from ...models import TableSchema
from .utils import refresh_dynamic
from ..types import DynamicSchema
from ...compat import six


class BaseAnalyzer(Backend):
    """
    Analyzer is used before optimzing,
    which analyze some operation that is not supported for this execution backend.
    """

    def __init__(self, expr_dag, traversed=None, on_sub=None):
        self._dag = expr_dag
        self._indexer = itertools.count(0)
        self._traversed = traversed or set()
        self._on_sub = on_sub

    def analyze(self):
        for node in self._iter():
            self._traversed.add(id(node))
            self._visit_node(node)

        return self._dag.root

    def _iter(self):
        for node in traverse_until_source(self._dag, top_down=True,
                                          traversed=self._traversed):
            yield node

        while True:
            all_traversed = True
            for node in traverse_until_source(self._dag, top_down=True):
                if id(node) not in self._traversed:
                    all_traversed = False
                    yield node
            if all_traversed:
                break

    def _visit_node(self, node):
        try:
            node.accept(self)
        except NotImplementedError:
            return

    def _sub(self, expr, sub, parents=None):
        self._dag.substitute(expr, sub, parents=parents)
        if self._on_sub:
            self._on_sub(expr, sub)

    @staticmethod
    def _get_moment_sub_expr(expr, _input, order, center):
        def _group_mean(e):
            m = e.mean()
            if isinstance(expr, GroupedSequenceReduction):
                m = m.to_grouped_reduction(expr._grouped)
            return m

        def _order(e, o):
            if o == 1:
                return e
            else:
                return e ** o

        if not center:
            if order == 0:
                sub = Scalar(1)
            else:
                sub = _group_mean(_input ** order)
        else:
            if order == 0:
                sub = Scalar(1)
            elif order == 1:
                sub = Scalar(0)
            else:
                sub = _group_mean(_input ** order)
                divided = 1
                divisor = 1
                for o in compat.irange(1, order):
                    divided *= order - o + 1
                    divisor *= o
                    part_item = divided // divisor * _group_mean(_order(_input, order - o)) \
                                * (_order(_group_mean(_input), o))
                    if o & 1:
                        sub -= part_item
                    else:
                        sub += part_item
                part_item = _group_mean(_input) ** order
                if order & 1:
                    sub -= part_item
                else:
                    sub += part_item
        return sub

    @classmethod
    def _get_cut_sub_expr(cls, expr):
        is_seq = isinstance(expr, SequenceExpr)
        kw = dict()
        if is_seq:
            kw['_data_type'] = expr.dtype
        else:
            kw['_value_type'] = expr.dtype

        conditions = []
        thens = []

        if expr.include_under:
            bin = expr.bins[0]
            if expr.right and not expr.include_lowest:
                conditions.append(expr.input <= bin)
            else:
                conditions.append(expr.input < bin)
            thens.append(expr.labels[0])
        for i, bin in enumerate(expr.bins[1:]):
            lower_bin = expr.bins[i]
            if not expr.right or (i == 0 and expr.include_lowest):
                condition = lower_bin <= expr.input
            else:
                condition = lower_bin < expr.input

            if expr.right:
                condition = (condition & (expr.input <= bin))
            else:
                condition = (condition & (expr.input < bin))

            conditions.append(condition)
            if expr.include_under:
                thens.append(expr.labels[i + 1])
            else:
                thens.append(expr.labels[i])
        if expr.include_over:
            bin = expr.bins[-1]
            if expr.right:
                conditions.append(bin < expr.input)
            else:
                conditions.append(bin <= expr.input)
            thens.append(expr.labels[-1])

        return Switch(_conditions=conditions, _thens=thens,
                      _default=None, _input=None, **kw)

    @classmethod
    def _get_value_counts_sub_expr(cls, expr):
        collection = expr.input
        by = expr._by
        sort = expr._sort.value
        ascending = expr._ascending.value
        dropna = expr._dropna.value

        sub = collection.groupby(by).agg(count=collection.count())
        if sort:
            sub = sub.sort('count', ascending=ascending)
        if dropna:
            sub = sub.filter(sub[by.name].notnull())

        return sub

    def _get_pivot_sub_expr(self, expr):
        columns_expr = expr.input.distinct([c.copy() for c in expr._columns])

        group_names = [g.name for g in expr._group]
        group_types = [g.dtype for g in expr._group]
        exprs = [expr]

        def callback(result, new_expr):
            expr = exprs[0]
            columns = [r[0] for r in result]

            if len(expr._values) > 1:
                names = group_names + \
                        ['{0}_{1}'.format(v.name, c)
                         for v in expr._values for c in columns]
                types = group_types + \
                        list(itertools.chain(*[[n.dtype] * len(columns)
                                               for n in expr._values]))
            else:
                names = group_names + columns
                types = group_types + [expr._values[0].dtype] * len(columns)
            new_expr._schema = TableSchema.from_lists(names, types)

            column_name = expr._columns[0].name  # column's size can only be 1
            values_names = [v.name for v in expr._values]

            @output(names, types)
            def reducer(keys):
                values = [None] * len(columns) * len(values_names)

                def h(row, done):
                    col = getattr(row, column_name)
                    for val_idx, value_name in enumerate(values_names):
                        val = getattr(row, value_name)
                        idx = len(columns) * val_idx + columns.index(col)
                        if values[idx] is not None:
                            raise ValueError(
                                'Row contains duplicate entries, rows: {0}, column: {1}'.format(keys, col))
                        values[idx] = val
                    if done:
                        yield keys + tuple(values)

                return h

            fields = expr._group + expr._columns + expr._values
            pivoted = expr.input.select(fields).map_reduce(reducer=reducer, group=group_names)
            self._sub(new_expr, pivoted)

            # trigger refresh of dynamic operations
            refresh_dynamic(pivoted, self._dag)

        return CollectionExpr(_schema=DynamicSchema.from_lists(group_names, group_types),
                              _deps=[(columns_expr, callback)])

    def _get_pivot_table_sub_expr_without_columns(self, expr):
        def get_agg(field, agg_func, agg_func_name, fill_value):
            from ..expr.expressions import ReprWrapper

            if isinstance(agg_func, six.string_types):
                aggregated = field.eval(agg_func, rewrite=False)
                if isinstance(aggregated, ReprWrapper):
                    aggregated = aggregated()
            else:
                aggregated = field.agg(agg_func)
            if fill_value is not None:
                aggregated.fillna(fill_value)
            return aggregated.rename('{0}_{1}'.format(field.name, agg_func_name))

        grouped = expr.input.groupby(expr._group)
        aggs = []
        for agg_func, agg_func_name in zip(expr._agg_func, expr._agg_func_names):
            for value in expr._values:
                agg = get_agg(value, agg_func, agg_func_name, expr.fill_value)
                aggs.append(agg)
        return grouped.aggregate(aggs, sort_by_name=False)

    def _get_pivot_table_sub_expr_with_columns(self, expr):
        columns_expr = expr.input.distinct([c.copy() for c in expr._columns])

        group_names = [g.name for g in expr._group]
        group_types = [g.dtype for g in expr._group]
        exprs = [expr]

        def callback(result, new_expr):
            expr = exprs[0]
            columns = [r[0] for r in result]

            names = list(group_names)
            tps = list(group_types)
            aggs = []
            for agg_func_name, agg_func in zip(expr._agg_func_names, expr._agg_func):
                for value_col in expr._values:
                    for col in columns:
                        base = '{0}_'.format(col) if col is not None else ''
                        name = '{0}{1}_{2}'.format(base, value_col.name, agg_func_name)
                        names.append(name)
                        tps.append(value_col.dtype)

                        col = col.item() if hasattr(col, 'item') else col
                        field = (expr._columns[0] == col).ifelse(
                            value_col, Scalar(_value_type=value_col.dtype))
                        if isinstance(agg_func, six.string_types):
                            agg = getattr(field, agg_func)()
                        else:
                            func = agg_func()

                            class ActualAgg(object):
                                def buffer(self):
                                    return func.buffer()

                                def __call__(self, buffer, value):
                                    if value is None:
                                        return
                                    func(buffer, value)

                                def merge(self, buffer, pbuffer):
                                    func.merge(buffer, pbuffer)

                                def getvalue(self, buffer):
                                    return func.getvalue(buffer)

                            agg = field.agg(ActualAgg)
                        if expr.fill_value is not None:
                            agg = agg.fillna(expr.fill_value)
                        agg = agg.rename(name)
                        aggs.append(agg)

            new_expr._schema = TableSchema.from_lists(names, tps)

            pivoted = expr.input.groupby(expr._group).aggregate(aggs, sort_by_name=False)
            self._sub(new_expr, pivoted)

            # trigger refresh of dynamic operations
            refresh_dynamic(pivoted, self._dag)

        return CollectionExpr(_schema=DynamicSchema.from_lists(group_names, group_types),
                              _deps=[(columns_expr, callback)])

    def _get_pivot_table_sub_expr(self, expr):
        if expr._columns is None:
            return self._get_pivot_table_sub_expr_without_columns(expr)
        else:
            return self._get_pivot_table_sub_expr_with_columns(expr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import requests

from ..compat import ElementTree as ET, TimeoutError
from ..errors import ODPSError


class TunnelError(ODPSError):

    @classmethod
    def parse(cls, resp):
        try:
            root = ET.fromstring(resp.content)
            code = root.find('./Code').text
            msg = root.find('./Message').text
            request_id = root.find('./RequestId')
            if request_id:
                request_id = request_id.text
            else:
                request_id = resp.headers.get('x-odps-request-id')
            exc_type = globals().get(code, TunnelError)
        except:
            request_id = resp.headers['x-odps-request-id']
            obj = json.loads(resp.content)
            msg = obj['Message']
            code = obj['InvalidArgument']
            exc_type = globals().get(code, TunnelError)

        return exc_type(msg, code=code, request_id=request_id)


class TunnelReadTimeout(TunnelError, TimeoutError, requests.ReadTimeout):
    pass


class TunnelWriteTimeout(TunnelError, TimeoutError, requests.ConnectionError):
    pass


class MetaTransactionFailed(TunnelError):
    pass



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import warnings

from .. import serializers
from ..compat import Enum
from ..errors import InvalidParameter, MethodNotAllowed
from ..utils import parse_rfc822, with_wait_argument
from .core import JSONRemoteModel, LazyLoad
from .functions import Functions
from .resources import Resources
from .tables import Tables
from .volumes import Volumes


def _parse_schema_time(time_str):
    try:
        return parse_rfc822(time_str, use_legacy_parsedate=False)
    except (TypeError, ValueError):
        return datetime.datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S")


class SchemaType(Enum):
    MANAGED = 0
    EXTERNAL = 1


class SchemaDescription(JSONRemoteModel):
    name = serializers.JSONNodeField('name', set_to_parent=True)
    owner = serializers.JSONNodeField('owner', set_to_parent=True)
    description = serializers.JSONNodeField('description', set_to_parent=True)
    creation_time = serializers.JSONNodeField(
        'createTime', parse_callback=_parse_schema_time, set_to_parent=True
    )
    last_modified_time = serializers.JSONNodeField(
        'modifyTime', parse_callback=_parse_schema_time, set_to_parent=True
    )
    type = serializers.JSONNodeField(
        'type', parse_callback=lambda x: getattr(SchemaType, x.upper())
    )


class Schema(LazyLoad):
    default_schema_name = "DEFAULT"

    _root = 'Schema'

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    description = serializers.XMLNodeField('Description')
    creation_time = serializers.XMLNodeField('CreateTime', parse_callback=_parse_schema_time)
    last_modified_time = serializers.XMLNodeField('ModifyTime', parse_callback=_parse_schema_time)
    type = serializers.XMLNodeField(
        'Type', parse_callback=lambda x: getattr(SchemaType, x.upper())
    )

    def reload(self):
        try:
            self.parent._check_schema_api()
            resp = self._client.get(self.resource() + "/schemas/" + self.name)
            self.parse(self._client, resp, obj=self)

            self.owner = resp.headers.get('x-odps-owner')
            self.creation_time = parse_rfc822(
                resp.headers.get('x-odps-creation-time'), use_legacy_parsedate=False
            )
            self.last_modified_time = parse_rfc822(
                resp.headers.get("Last-Modified"), use_legacy_parsedate=False
            )
            self._loaded = True
        except (InvalidParameter, MethodNotAllowed):
            desc_instance = self.project.odps.execute_sql("DESC SCHEMA %s" % self.name)
            desc_result = desc_instance.get_task_results().get("AnonymousSQLTask")
            desc_obj = SchemaDescription(parent=self)
            desc_obj.parse(self._client, desc_result, obj=desc_obj)
            self._loaded = True

    @property
    def create_time(self):
        warnings.warn(
            'Schema.create_time is deprecated and will be replaced '
            'by Schema.creation_time.',
            DeprecationWarning,
            stacklevel=3,
        )
        return self.creation_time

    @property
    def modify_time(self):
        warnings.warn(
            'Schema.modify_time is deprecated and will be replaced '
            'by Schema.last_modified_time.',
            DeprecationWarning,
            stacklevel=3,
        )
        return self.last_modified_time

    def resource(self, client=None, endpoint=None):
        return self.parent.resource(client, endpoint=endpoint)

    @with_wait_argument
    def drop(self, async_=False):
        self.parent.delete(self, async_=async_)

    @property
    def functions(self):
        return Functions(client=self._client, parent=self)

    @property
    def resources(self):
        return Resources(client=self._client, parent=self)

    @property
    def tables(self):
        return Tables(client=self._client, parent=self)

    @property
    def volumes(self):
        return Volumes(client=self._client, parent=self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
import random

from ...models import TableSchema
from .expressions import Expr, CollectionExpr, BooleanSequenceExpr, \
    Column, SequenceExpr, Scalar, BooleanScalar, repr_obj
from .collections import SortedExpr, ReshuffledCollectionExpr
from .errors import ExpressionError
from . import utils
from ...compat import reduce, six
from .. import types
from ..utils import is_constant_scalar
from ...utils import object_getattr, camel_to_underline


class BaseGroupBy(Expr):
    __slots__ = '_to_agg', '_by_names'
    _args = '_input', '_by'

    def _init(self, *args, **kwargs):
        self._init_attr('_to_agg', None)
        super(BaseGroupBy, self)._init(*args, **kwargs)
        if isinstance(self._by, list):
            self._by = self._input._get_fields(self._by)
        else:
            self._by = [self._input._get_field(self._by)]
        for idx, by_field in enumerate(self._by):
            if by_field.name is None:
                new_field_name = '_%s_%d' % (camel_to_underline(type(by_field).__name__),
                                             random.randint(10000, 99999))
                self._by[idx] = by_field.rename(new_field_name)
        if self._to_agg is None:
            self._to_agg = self._input.schema

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            if item in self._to_agg:
                return SequenceGroupBy(_input=self, _name=item,
                                       _data_type=self._input._schema[item].type)
            else:
                raise KeyError('Fail to get group by field, unknown field: %s' % repr_obj(item))

        is_field = lambda it: isinstance(it, six.string_types) or isinstance(it, Column)
        if not all(is_field(it) for it in item):
            raise TypeError('Fail to get group by fields, unknown type: %s' % type(item))
        if any(col.is_renamed() for col in item if isinstance(col, Column)):
            raise ValueError('Fail to get group by fields, column cannot be renamed')

        get_name = lambda it: it if isinstance(it, six.string_types) else it.source_name
        _to_agg = type(self._input.schema)(
            columns=self._input.schema[[get_name(field) for field in item
                                        if get_name(field) in self._to_agg]])

        return GroupBy(_input=self._input, _by=self._by,
                       _by_names=getattr(self, '_by_names', None), _to_agg=_to_agg)

    def __getattr__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError as e:
            if attr.startswith('__'):
                raise e

            agg = object.__getattribute__(self, '_to_agg')
            if agg is not None and attr in agg:
                return self[attr]

            raise e

    def sort_values(self, by, ascending=True):
        if hasattr(self, '_having') and self._having is not None:
            raise ExpressionError('Cannot sort GroupBy with `having`')

        if not isinstance(by, list):
            by = [by, ]
        by = [self._defunc(it) for it in by]

        attr_values = dict((attr, object_getattr(self, attr, None))
                           for attr in utils.get_attrs(self))
        attr_values['_sorted_fields'] = by
        attr_values['_ascending'] = ascending
        attr_values.pop('_having', None)

        return SortedGroupBy(**attr_values)

    def sort(self, *args, **kwargs):
        return self.sort_values(*args, **kwargs)

    def mutate(self, *windows, **kw):
        if hasattr(self, '_having') and self._having is not None:
            raise ExpressionError('Cannot mutate GroupBy with `having`')

        if len(windows) == 1 and isinstance(windows[0], list):
            windows = windows[0]
        else:
            windows = list(windows)
        windows = [self._defunc(win) for win in windows]
        if kw:
            windows.extend([self._defunc(win).rename(new_name)
                            for new_name, win in six.iteritems(kw)])

        from .window import Window

        if not windows:
            raise ValueError('Cannot mutate on grouped data')
        if not all(isinstance(win, Window) for win in windows):
            raise TypeError('Only window functions can be provided')

        win_field_names = filter(lambda it: it is not None,
                                 [win.source_name for win in windows])
        if not frozenset(win_field_names).issubset(self._to_agg.names):
            for agg_field_name in win_field_names:
                if agg_field_name not in self._to_agg:
                    raise ValueError('Unknown field to aggregate: %s' % repr_obj(agg_field_name))

        names = [by.name for by in self._by if isinstance(by, Column)] + \
                [win.name for win in windows]
        types = [by._data_type for by in self._by if isinstance(by, Column)] + \
                [win._data_type for win in windows]

        return MutateCollectionExpr(
            _input=self,
            _window_fields=windows,
            _schema=TableSchema.from_lists(names, types)
        )

    def apply(self, func, names=None, types=None, resources=None, args=(), **kwargs):
        reshuffled = ReshuffledCollectionExpr(_input=self, _schema=self._input._schema)
        return reshuffled.apply(axis=1, func=func, names=names, types=types,
                                resources=resources, args=args, **kwargs)


class GroupBy(BaseGroupBy):
    __slots__ = '_having',

    def _init(self, *args, **kwargs):
        self._init_attr('_having', None)
        super(GroupBy, self)._init(*args, **kwargs)

    def _same_by(self, other):
        if other._input is not self._input:
            return False
        if len(self._by) != len(other._by):
            return False
        if any(x is not y for x, y in zip(self._by, other._by)):
            return False
        return True

    def _validate_agg(self, agg):
        from .reduction import GroupedSequenceReduction
        from .window import RankOp

        has_reduction = False
        for node in agg.traverse(top_down=True, unique=True,
                                 stop_cond=lambda x: x is self._input):
            if isinstance(node, GroupedSequenceReduction):
                has_reduction = True
                if not self._same_by(node._grouped):
                    raise ExpressionError(
                        'Aggregation has not been applied to the right GroupBy, got: %s' % repr_obj(agg))
            elif isinstance(node, Column):
                if node._input is not self._input:
                    raise ExpressionError(
                        'Aggregation should be applied to the column of %s' % repr_obj(self._input))
            elif isinstance(node, RankOp) and node._input is not self._input:
                raise ExpressionError(
                    'Aggregation should be applied to the column of %s' % repr_obj(self._input))

        if not has_reduction:
            raise ExpressionError('No aggregation found in %s' % repr_obj(agg))

    def _transform(self, reduction_expr):
        if isinstance(reduction_expr, Scalar):
            from .reduction import SequenceReduction
            from .window import RankOp

            dag = reduction_expr.to_dag(copy=False, validate=False)
            for node in dag.traverse(
                    stop_cond=lambda x: isinstance(x, (Column, RankOp)) or x is self._input):
                if isinstance(node, SequenceReduction):
                    to_sub = node.to_grouped_reduction(self)
                    dag.substitute(node, to_sub)
                elif isinstance(node, Scalar) and not is_constant_scalar(node) \
                        and len(node.children()) > 0:
                    to_sub = node.to_sequence()
                    dag.substitute(node, to_sub)
                elif isinstance(node, Column) and node._input is not self._input:
                    field = self._input._get_field(node)
                    if field:
                        dag.substitute(node, field)
                elif isinstance(node, RankOp) and node._input is not self._input:
                    dag.substitute(node._input, self._input, parents=[node])

            return dag.root

        return reduction_expr

    def __repr__(self):
        return object.__repr__(self)

    def __getitem__(self, item):
        item = self._defunc(item)

        if isinstance(item, (BooleanSequenceExpr, BooleanScalar)):
            having = item if isinstance(item, BooleanSequenceExpr) \
                else self._transform(item)
            if self._having is not None:
                having = having & self._having
            return GroupBy(_input=self._input, _by=self._by,
                           _to_agg=self._to_agg,
                           _having=having)
        else:
            return super(GroupBy, self).__getitem__(item)

    def filter(self, *predicates):
        predicates = [self._defunc(it) for it in predicates]

        predicate = reduce(operator.and_, predicates)
        return self[predicate]

    def aggregate(self, *aggregations, **kw):
        sort_by_name = kw.pop('sort_by_name', True)

        if len(aggregations) == 1 and isinstance(aggregations[0], list):
            aggregations = aggregations[0]
        else:
            aggregations = list(aggregations)
        aggregations = [self._defunc(it) for it in aggregations]
        if kw:
            aggregations.extend([self._defunc(agg).rename(new_name)
                                 for new_name, agg in six.iteritems(kw)])

        # keep sequence to ensure that test cases work well
        if sort_by_name:
            aggregations = sorted([self._transform(agg) for agg in aggregations],
                                  key=lambda it: it.name)
        else:
            aggregations = [self._transform(agg) for agg in aggregations]

        if not aggregations:
            raise ValueError('Cannot aggregate on grouped data')
        [self._validate_agg(agg) for agg in aggregations]

        names = [by.name for by in self._by
                 if isinstance(by, (Scalar, SequenceExpr)) and by.name is not None] + \
                [agg.name for agg in aggregations]
        types = [by.dtype for by in self._by
                 if isinstance(by, (Scalar, SequenceExpr)) and by.name is not None] + \
                [agg._data_type for agg in aggregations]

        return GroupByCollectionExpr(
            _input=self,
            _aggregations=aggregations,
            _schema=TableSchema.from_lists(names, types)
        )

    def agg(self, *args, **kwargs):
        return self.aggregate(*args, **kwargs)


class SequenceGroupBy(Expr):
    __slots__ = '_name', '_data_type', '_source_data_type'
    _args = '_input',

    def _init(self, *args, **kwargs):
        self._init_attr('_data_type', None)
        self._init_attr('_source_data_type', None)
        super(SequenceGroupBy, self)._init(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        data_type = kwargs.get('_data_type')
        if data_type:
            cls_name = data_type.__class__.__name__ + SequenceGroupBy.__name__
            clazz = globals()[cls_name]

            return super(SequenceGroupBy, clazz).__new__(clazz)
        else:
            return super(SequenceGroupBy, cls).__new__(cls)

    @property
    def name(self):
        return self._name

    @property
    def dtype(self):
        return self._data_type

    @property
    def input(self):
        return self._input

    def astype(self, data_type):
        data_type = types.validate_data_type(data_type)

        if data_type == self._data_type:
            return self

        attr_dict = dict((k, getattr(self, k, None)) for k in utils.get_attrs(self))
        attr_dict['_data_type'] = data_type
        attr_dict['_source_data_type'] = self._source_data_type

        cls = globals().get(repr(data_type).capitalize() + SequenceGroupBy.__name__)
        new_sequence_groupby = cls(**attr_dict)

        return new_sequence_groupby

    def is_astyped(self):
        if self._source_data_type is None:
            return False
        return self._data_type != self._source_data_type

    def to_column(self):
        collection = self.input.input
        input = collection[self.name]
        if self.is_astyped():
            input = input.astype(self._data_type)
        return input


class BooleanSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(BooleanSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.boolean


class Int8SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Int8SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.int8


class Int16SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Int16SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.int16


class Int32SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Int32SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.int32


class Int64SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Int64SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.int64


class Float32SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Float32SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.float32


class Float64SequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(Float64SequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.float64


class DecimalSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(DecimalSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.decimal


class StringSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(StringSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.string


class BinarySequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(BinarySequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.binary


class DatetimeSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(DatetimeSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.datetime


class DateSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(DateSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.datetime


class TimestampSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(TimestampSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.datetime


class UnknownSequenceGroupBy(SequenceGroupBy):
    def _init(self, *args, **kwargs):
        super(UnknownSequenceGroupBy, self)._init(*args, **kwargs)
        self._data_type = types.Unknown()


class SortedGroupBy(BaseGroupBy, SortedExpr):
    __slots__ = '_sorted_fields', '_ascending'


class GroupByCollectionExpr(CollectionExpr):
    _args = '_input', '_by', '_having', '_aggregations', '_fields'
    node_name = 'GroupBy'

    def _init(self, *args, **kwargs):
        self._init_attr('_fields', None)

        super(GroupByCollectionExpr, self)._init(*args, **kwargs)

        if isinstance(self._input, GroupBy):
            self._by = self._input._by
            self._having = self._input._having
            self._input = self._input._input

    def iter_args(self):
        arg_names = ['collection', 'bys', 'having', 'aggregations']
        for it in zip(arg_names, self.args):
            yield it
        if self._fields is not None:
            yield ('selections', self._fields)

    def _name_to_exprs(self):
        if hasattr(self, '_fields') and self._fields is not None:
            exprs = self._fields
        else:
            exprs = self.args[1] + self.args[3]
        return dict((expr.name, expr) for expr in exprs if hasattr(expr, 'name'))

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_groupby(self)

    @property
    def fields(self):
        if self._fields is not None:
            return self._fields

        return self._by + self._aggregations


class MutateCollectionExpr(CollectionExpr):
    _args = '_input', '_by', '_window_fields', '_fields'
    node_name = 'Mutate'

    def _init(self, *args, **kwargs):
        self._init_attr('_fields', None)
        super(MutateCollectionExpr, self)._init(*args, **kwargs)

        if isinstance(self._input, GroupBy):
            self._by = self._input._by
            self._input = self._input._input

    @property
    def _project_fields(self):
        return self._window_fields

    def iter_args(self):
        for it in zip(['collection', 'bys', 'mutates'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    @property
    def fields(self):
        if self._fields is not None:
            return self._fields

        return self._by + self._window_fields

    def accept(self, visitor):
        return visitor.visit_mutate(self)


def groupby(expr, by, *bys):
    """
    Group collection by a series of sequences.

    :param expr: collection
    :param by: columns to group
    :param bys: columns to group
    :return: GroupBy instance
    :rtype: :class:`odps.df.expr.groupby.GroupBy`
    """

    if not isinstance(by, list):
        by = [by, ]
    if len(bys) > 0:
        by = by + list(bys)
    return GroupBy(_input=expr, _by=by)


CollectionExpr.groupby = groupby


class ValueCounts(CollectionExpr):
    _args = '_input', '_by', '_sort', '_ascending', '_dropna'
    node_name = 'ValueCounts'

    def _init(self, *args, **kwargs):
        super(ValueCounts, self)._init(*args, **kwargs)

        if isinstance(self._input, SequenceExpr):
            self._by = self._input
            self._input = next(it for it in self._input.traverse(top_down=True)
                               if isinstance(it, CollectionExpr))
        if isinstance(self._sort, bool):
            self._sort = Scalar(_value=self._sort)
        if isinstance(self._ascending, bool):
            self._ascending = Scalar(_value=self._ascending)
        if isinstance(self._dropna, bool):
            self._dropna = Scalar(_value=self._dropna)

    def iter_args(self):
        for it in zip(['collection', 'by', 'sort', 'ascending', 'dropna'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_value_counts(self)


def value_counts(expr, sort=True, ascending=False, dropna=False):
    """
    Return object containing counts of unique values.

    The resulting object will be in descending order so that the first element is the most frequently-occuring
    element. Exclude NA values by default

    :param expr: sequence
    :param sort: if sort
    :type sort: bool
    :param dropna: Don’t include counts of None, default False
    :return: collection with two columns
    :rtype: :class:`odps.df.expr.expressions.CollectionExpr`
    """

    names = [expr.name, 'count']
    typos = [expr.dtype, types.int64]
    return ValueCounts(_input=expr, _schema=TableSchema.from_lists(names, typos),
                       _sort=sort, _ascending=ascending, _dropna=dropna)


def topk(expr, k):
    return expr.value_counts().limit(k)


SequenceExpr.value_counts = value_counts
SequenceExpr.topk = topk
